# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""Tests for interface for getting Crash and Bucket data from CrashManager API"""
import json
from zipfile import ZipFile

from FTB.ProgramConfiguration import ProgramConfiguration
from pytest import mark, raises

from .fuzzmanager import Bucket, CrashEntry, load_fm_data
from .storage import TEST_INFO


def test_bucket_01(mocker):
    """bucket getattr uses data from get"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.get.return_value.json.return_value = {"testcase": "data"}
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    bucket = Bucket(123)
    assert coll.return_value.get.call_count == 0
    assert bucket.testcase == "data"
    with raises(AttributeError):
        getattr(bucket, "other")
    assert coll.return_value.get.call_count == 1


def test_bucket_02(mocker):
    """bucket setattr raises"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    bucket = Bucket(123)
    with raises(AttributeError):
        bucket.other = "data"
    assert coll.return_value.get.call_count == 0


def test_bucket_03(mocker):
    """bucket iter_crashes flattens across pages"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.side_effect = [
        {
            "count": 2,
            "next": "url",
            "results": [
                {"id": 234, "testcase": "test1"},
                {"id": 345, "testcase": None},
            ],
        },
        {
            "count": 1,
            "next": None,
            "results": [
                {"id": 456, "testcase": "test2"},
            ],
        },
    ]
    bucket = Bucket(123)
    assert coll.return_value.get.call_count == 0
    crashes = list(bucket.iter_crashes(quality_filter=5))
    assert coll.return_value.get.call_count == 2
    assert coll.return_value.get.call_args_list[0][1]["params"]["include_raw"] == "0"
    assert json.loads(
        coll.return_value.get.call_args_list[0][1]["params"]["query"]
    ) == {
        "op": "AND",
        "bucket": 123,
        "testcase__quality": 5,
    }
    assert len(crashes) == 2
    assert crashes[0].crash_id == 234
    assert crashes[1].crash_id == 456


def test_bucket_04(mocker):
    """bucket signature_path writes and returns sig json and metadata"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "signature": "sigdata",
        "size": 10,
        "frequent": True,
        "shortDescription": "sig desc",
        "best_quality": 0,
    }
    with Bucket(123) as bucket:
        assert coll.return_value.get.call_count == 0
        sig_path = bucket.signature_path()
        assert sig_path.is_file()
        assert sig_path.with_suffix(".metadata").is_file()
        assert sig_path.read_text() == "sigdata"
        assert json.loads(sig_path.with_suffix(".metadata").read_text()) == {
            "size": 10,
            "frequent": True,
            "shortDescription": "sig desc",
            "testcase__quality": 0,
        }
        assert coll.return_value.get.call_count == 1
        # second call returns same path
        assert bucket.signature_path() == sig_path
    assert coll.return_value.get.call_count == 1


def test_crash_01(mocker):
    """crash getattr uses data from get"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.get.return_value.json.return_value = {"testcase": "data"}
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    crash = CrashEntry(123)
    assert coll.return_value.get.call_count == 0
    assert crash.testcase == "data"
    with raises(AttributeError):
        getattr(crash, "other")
    assert coll.return_value.get.call_count == 1
    assert coll.return_value.get.call_args[1]["params"] == {"include_raw": "0"}

    # crash getattr for raw field re-gets
    coll.return_value.get.return_value.json.return_value = {"rawStderr": "stderr"}
    assert crash.rawStderr == "stderr"
    assert coll.return_value.get.call_count == 2
    assert coll.return_value.get.call_args[1]["params"] == {"include_raw": "1"}


def test_crash_02(mocker):
    """crash setattr raises except testcase_quality"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.get.return_value.json.return_value = {"testcase": "data"}
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    crash = CrashEntry(123)

    # crash setattr raises for other field
    with raises(AttributeError):
        crash.other = "data"
    assert coll.return_value.get.call_count == 0

    # crash setattr for testcase_quality works and updates data if set
    assert coll.return_value.patch.call_count == 0
    crash.testcase_quality = 5
    assert coll.return_value.get.call_count == 0
    assert coll.return_value.patch.call_count == 1
    with raises(AttributeError):
        getattr(crash, "testcase_quality")
    assert coll.return_value.get.call_count == 1
    getattr(crash, "testcase")
    assert coll.return_value.get.call_count == 1
    crash.testcase_quality = 10
    assert coll.return_value.patch.call_count == 2
    assert crash.testcase_quality == 10
    assert coll.return_value.get.call_count == 1


@mark.parametrize(
    "file_name, passed_ext, expected",
    [
        # using existing name
        ("foo.html", None, "test.html"),
        # use default extension
        ("foo", None, "test.html"),
        # use default extension
        ("foo.", None, "test.html"),
        # add missing extension
        ("foo", "svg", "test.svg"),
        # overwrite extension
        ("foo.zip", "svg", "test.svg"),
        # bad zipfile
        ("foo.zip", None, None),
    ],
)
def test_crash_03(mocker, tmp_path, file_name, passed_ext, expected):
    """test case is not zipfile"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "id": 234,
        "testcase": file_name,
    }
    with (tmp_path / file_name).open("w") as zip_fp:
        zip_fp.write("data")
    with CrashEntry(234) as crash:
        assert crash.testcase == file_name  # pre-load data dict so I can re-patch get
        coll.return_value.get.return_value = mocker.Mock(
            content=(tmp_path / file_name).read_bytes(),
            headers={"content-disposition": f'attachment; filename="bar/{file_name}"'},
        )
        assert coll.return_value.get.call_count == 1
        tests = crash.testcases(ext=passed_ext)
        assert crash._contents is not None
        if expected is not None:
            assert tests
            assert tests[-1].name == expected
        else:
            assert not tests


def test_crash_04(mocker, tmp_path):
    """crash testcase_path writes and returns testcase zip"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "id": 234,
        "testcase": "test.zip",
    }
    # build archive containing test cases
    with ZipFile(tmp_path / "test.zip", "w") as zip_fp:
        # add files out of order
        for i in (1, 3, 2, 0):
            test = tmp_path / f"test-{i}"
            test.mkdir()
            (test / TEST_INFO).touch()
            zip_fp.write(test / TEST_INFO, arcname=f"test-{i}/{TEST_INFO}")
    with CrashEntry(234) as crash:
        assert crash.testcase == "test.zip"  # pre-load data dict so I can re-patch get
        coll.return_value.get.return_value = mocker.Mock(
            content=(tmp_path / "test.zip").read_bytes(),
            headers={"content-disposition": 'attachment; filename="tests/test.zip"'},
        )
        assert coll.return_value.get.call_count == 1
        tests = crash.testcases()
        assert len(tests) == 4
        # check order
        assert tests[0].name == "test-3"
        assert tests[1].name == "test-2"
        assert tests[2].name == "test-1"
        assert tests[3].name == "test-0"
        assert coll.return_value.get.call_count == 2
        # second call returns same path
        assert crash.testcases() == tests
        # subsets
        assert crash.testcases(subset=[0]) == [tests[0]]
        # remove second oldest test case (oldest = 0, most recent = n-1)
        tests = crash.testcases(subset=[0, 2, 3])
        assert len(tests) == 3
        assert tests[0].name == "test-3"
        assert tests[1].name == "test-1"
        assert tests[2].name == "test-0"
    assert coll.return_value.get.call_count == 2


def test_crash_05(mocker):
    """crash create_signature writes and returns signature path"""
    cfg = ProgramConfiguration("product", "platform", "os")
    mocker.patch(
        "grizzly.common.fuzzmanager.ProgramConfiguration"
    ).fromBinary.return_value = cfg
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "rawStdout": "",
        "rawStderr": "",
        "rawCrashData": "ERROR: AddressSanitizer: SEGV on address 0x14 "
        "(pc 0x123 sp 0x456 bp 0x789 T0)",
    }
    with CrashEntry(123) as crash:
        assert coll.return_value.get.call_count == 0
        sig_path = crash.create_signature(None)
        assert sig_path.is_file()
        assert sig_path.with_suffix(".metadata").is_file()
        assert "AddressSanitizer" in sig_path.read_text()
        assert json.loads(sig_path.with_suffix(".metadata").read_text()) == {
            "size": 1,
            "frequent": False,
            "shortDescription": "AddressSanitizer: SEGV",
            "testcase__quality": 5,
        }
        assert coll.return_value.get.call_count == 1
        # second call returns same path
        assert crash.create_signature(None) == sig_path
    assert coll.return_value.get.call_count == 1


def test_crash_06(mocker):
    """crash create_signature raises when it can't create a signature"""
    cfg = ProgramConfiguration("product", "platform", "os")
    mocker.patch(
        "grizzly.common.fuzzmanager.ProgramConfiguration"
    ).fromBinary.return_value = cfg
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value.json.return_value = {
        "rawStdout": "",
        "rawStderr": "",
        "rawCrashData": "",
    }
    with raises(RuntimeError) as exc, CrashEntry(123) as crash:
        crash.create_signature(None)
    assert "insufficient data to generate" in str(exc).lower()
    assert coll.return_value.get.call_count == 1


def test_crash_07(tmp_path):
    """test CrashEntry._subset()"""
    # test single entry
    paths = [tmp_path / "0"]
    assert CrashEntry._subset(paths.copy(), [0]) == paths
    assert CrashEntry._subset(paths.copy(), [-1]) == paths
    # out of range (should be min/max'ed)
    assert CrashEntry._subset(paths.copy(), [1]) == paths
    # duplicate index
    assert CrashEntry._subset(paths.copy(), [1, 1]) == paths
    # test multiple entries select single
    paths = [tmp_path / "0", tmp_path / "1", tmp_path / "2"]
    assert CrashEntry._subset(paths.copy(), [0]) == [paths[0]]
    assert CrashEntry._subset(paths.copy(), [1]) == [paths[1]]
    assert CrashEntry._subset(paths.copy(), [2]) == [paths[2]]
    # out of range (should be min/max'ed)
    assert CrashEntry._subset(paths.copy(), [3]) == [paths[2]]
    assert CrashEntry._subset(paths.copy(), [-3]) == [paths[0]]
    # test multiple entries select multiple
    assert CrashEntry._subset(paths.copy(), [0, 1]) == paths[:2]
    assert CrashEntry._subset(paths.copy(), [2, 1]) == paths[1:]
    assert CrashEntry._subset(paths.copy(), [0, -1]) == [paths[0], paths[-1]]
    assert CrashEntry._subset(paths.copy(), [0, 1, -1]) == paths


@mark.parametrize(
    "bucket_id, load_bucket",
    [
        # Nothing to load, don't try
        (None, False),
        # Nothing to load, try
        (None, True),
        # Bucket exists, don't load it
        (111, False),
        # Bucket exists, load it
        (111, True),
    ],
)
def test_load_fm_data_01(mocker, bucket_id, load_bucket):
    """test load_fm_data()"""
    coll = mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    coll.return_value.serverProtocol = "http"
    coll.return_value.serverPort = 123
    coll.return_value.serverHost = "allizom.org"
    coll.return_value.get.return_value = mocker.Mock(
        content=b"\x01\x02\x03",
        headers={"content-disposition"},
    )
    coll.return_value.get.return_value.json.return_value = {"bucket": bucket_id}

    with load_fm_data(123, load_bucket) as (crash, bucket):
        assert isinstance(crash, CrashEntry)
        if load_bucket and bucket_id:
            assert isinstance(bucket, Bucket)
        else:
            assert bucket is None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly status"""
# pylint: disable=protected-access

from contextlib import closing
from dataclasses import astuple, fields
from itertools import count
from multiprocessing import Event, Process
from sqlite3 import connect
from time import sleep

from pytest import mark

from .reporter import FuzzManagerReporter
from .status import (
    DB_VERSION,
    REPORT_RATE,
    BaseStatus,
    ReadOnlyResultCounter,
    ReadOnlyStatus,
    ReductionStatus,
    ReductionStep,
    ResultCounter,
    ResultEntry,
    SimpleResultCounter,
    SimpleStatus,
    Status,
    _db_version_check,
)


def test_basic_status_01():
    """test BaseStatus()"""
    status = BaseStatus(1337, 1234567.89)
    assert status is not None
    assert status.pid == 1337
    assert status.start_time > 0
    assert status.ignored == 0
    assert status.iteration == 0
    assert status.log_size == 0
    assert not status._profiles
    assert status.runtime > 0
    assert status.rate == 0
    assert not any(status.profile_entries())


def test_basic_status_02(mocker):
    """test BaseStatus.runtime and BaseStatus.rate calculations"""
    fake_time = mocker.patch("grizzly.common.status.time", autospec=True)
    fake_time.return_value = 1.0
    status = BaseStatus(1337, 1.0)
    assert status.start_time == 1
    # test no iterations
    fake_time.return_value = 3.0
    assert status.runtime == 2.0
    assert status.rate == 0
    # test one iteration
    status.iteration = 1
    fake_time.return_value = 5.0
    assert status.runtime == 4.0
    assert status.rate == 0.25


def test_readonly_status_01():
    """test ReadOnlyStatus.runtime and ReadOnlyStatus.rate calculations"""
    # load from empty db
    status = ReadOnlyStatus(123, 1.0, 5.0)
    assert status.start_time == 1
    assert status.timestamp == 5
    assert status.runtime == 4.0
    assert status.iteration == 0
    assert status.rate == 0
    status.iteration = 1
    assert status.rate == 0.25


def test_simple_status_01():
    """test SimpleStatus()"""
    status = SimpleStatus.start()
    assert status is not None
    assert status.pid is not None
    assert status.start_time > 0
    assert status.results is not None


def test_status_01(tmp_path):
    """test Status.start()"""
    status = Status.start(tmp_path / "status.db")
    assert status is not None
    assert status._db_file is not None
    assert status.pid is not None
    assert status.start_time > 0
    assert status.timestamp >= status.start_time
    assert status.ignored == 0
    assert status.iteration == 0
    assert status.log_size == 0
    assert status.results is not None
    assert not status._enable_profiling
    assert not status._profiles


@mark.parametrize(
    "cur_time, force, reported, found_result",
    [
        # attempt report before REPORT_RATE elapses
        (count(start=1.0, step=0.1), False, False, False),
        # force report before REPORT_RATE elapses
        (count(start=1.0, step=0.1), True, True, False),
        # attempt report before REPORT_RATE elapses
        (count(start=1.0, step=0.1), False, True, True),
        # attempt report after REPORT_RATE elapses
        (count(start=1.0, step=REPORT_RATE + 1.0), False, True, False),
    ],
)
def test_status_02(mocker, tmp_path, cur_time, force, reported, found_result):
    """test Status.report()"""
    mocker.patch("grizzly.common.status.time", autospec=True, side_effect=cur_time)
    status = Status.start(tmp_path / "status.db")
    if found_result:
        status.results.count("uid1", "sig1")
    assert status.report(force=force) == reported


def test_status_03(tmp_path):
    """test ReadOnlyStatus.load_all()"""
    db_file = tmp_path / "status.db"
    # load from empty db
    assert not any(ReadOnlyStatus.load_all(db_file))
    # create simple entry
    status = Status.start(db_file, enable_profiling=True)
    status.results.count("uid1", "sig1")
    status.record("test", 123.45)
    status.report(force=True)
    assert status.results.total == 1
    loaded = next(ReadOnlyStatus.load_all(db_file))
    assert status.start_time == loaded.start_time
    assert status.timestamp == loaded.timestamp
    assert status.runtime >= loaded.runtime
    assert status.ignored == loaded.ignored
    assert status.iteration == loaded.iteration
    assert status.log_size == loaded.log_size
    assert status.pid == loaded.pid
    assert loaded.results.get("uid1") == ResultEntry("uid1", 1, "sig1")
    assert "test" in loaded._profiles


def test_status_04(mocker, tmp_path):
    """test ReadOnlyStatus.load_all() - multiple entries"""
    getpid = mocker.patch("grizzly.common.status.getpid", autospec=True)
    db_file = tmp_path / "status.db"
    for pid in range(5):
        getpid.return_value = pid
        Status.start(db_file)
    assert len(tuple(ReadOnlyStatus.load_all(db_file))) == 5


def test_status_05(mocker, tmp_path):
    """test ReadOnlyStatus.load_all() - filter entries by time"""
    fake_time = mocker.patch("grizzly.common.status.time", autospec=True)
    fake_time.return_value = 1.0
    db_file = tmp_path / "status.db"
    # create entry
    status = Status.start(db_file)
    status.results.count("uid1", "sig1")
    assert status.results.total == 1
    status.report(force=True)
    # load entry
    assert any(ReadOnlyStatus.load_all(db_file, time_limit=60))
    # load with expired entry
    fake_time.return_value = 1200.0
    assert not any(ReadOnlyStatus.load_all(db_file, time_limit=60))
    # load with no limit
    assert any(ReadOnlyStatus.load_all(db_file, time_limit=0))
    # load long running entry with a one month old result
    fake_time.return_value = 2592000.0
    status.report(force=True)
    loaded = next(ReadOnlyStatus.load_all(db_file, time_limit=60))
    assert status.start_time == loaded.start_time
    assert status.timestamp == loaded.timestamp
    assert status.runtime >= loaded.runtime
    assert status.ignored == loaded.ignored
    assert status.iteration == loaded.iteration
    assert status.log_size == loaded.log_size
    assert status.pid == loaded.pid
    assert loaded.results.get("uid1") == ResultEntry("uid1", 1, "sig1")


# NOTE: this function must be at the top level to work on Windows
def _client_writer(db_file, begin, iters):
    """Used by test_status_06"""
    begin.wait(timeout=45)
    status = Status.start(db_file)
    for _ in range(iters):
        status.iteration += 1
        status.report(force=True)
        sleep(0.01)


@mark.parametrize(
    "loads_in_parallel",
    [
        # only test reporting in parallel
        0,
        # test reporting and loading in parallel
        5,
    ],
)
def test_status_06(tmp_path, loads_in_parallel):
    """test ReadOnlyStatus.load_all() with multiple active clients in parallel"""
    begin = Event()
    clients = 10
    db_file = tmp_path / "status.db"
    iter_count = 5
    procs = []
    try:
        # create and launch client processes
        for _ in range(clients):
            procs.append(
                Process(target=_client_writer, args=(db_file, begin, iter_count))
            )
            procs[-1].start()
        # synchronize client processes (not perfect but good enough)
        begin.set()
        # attempt parallel loads
        for _ in range(loads_in_parallel):
            tuple(ReadOnlyStatus.load_all(db_file))
        # wait for processes to report and exit
        for proc in procs:
            proc.join(timeout=60)
            assert proc.exitcode == 0
        # collect reports
        reports = tuple(ReadOnlyStatus.load_all(db_file))
        # check that each process created a report
        assert len(reports) == clients
        # check reported data
        assert max(x.rate for x in reports) > 0
        assert sum(x.iteration for x in reports) == iter_count * clients
    finally:
        for proc in procs:
            if proc.exitcode is None:
                proc.terminate()
            proc.join()


def test_status_07(tmp_path):
    """test Status.measure() and Status.record() - profiling support"""
    db_file = tmp_path / "status.db"
    # profiling disabled
    status = Status.start(db_file, enable_profiling=False)
    status.record("x", 10.1)
    assert not status._profiles
    with status.measure("x"):
        pass
    assert not status._profiles
    # profiling enabled
    status = Status.start(db_file, enable_profiling=True)
    assert not status._profiles
    # initial entry
    status.record("test1", 10.1)
    assert "test1" in status._profiles
    assert status._profiles["test1"]["count"] == 1
    assert status._profiles["test1"]["max"] == 10.1
    assert status._profiles["test1"]["min"] == 10.1
    assert status._profiles["test1"]["total"] == 10.1
    entry = next(status.profile_entries())
    assert entry.name == "test1"
    assert entry.count == 1
    assert entry.max == 10.1
    assert entry.min == 10.1
    assert entry.total == 10.1
    # new min
    status.record("test1", 0.4)
    entry = next(status.profile_entries())
    assert entry.count == 2
    assert entry.max == 10.1
    assert entry.min == 0.4
    assert entry.total == 10.5
    # entry
    status.record("test1", 2)
    entry = next(status.profile_entries())
    assert entry.count == 3
    assert entry.max == 10.1
    assert entry.min == 0.4
    assert entry.total == 12.5
    # new max
    status.record("test1", 99.12)
    entry = next(status.profile_entries())
    assert entry.count == 4
    assert entry.max == 99.12
    assert entry.min == 0.4
    assert entry.total == 111.62
    # new name
    status.record("test2", 1)
    assert "test2" in status._profiles
    assert len(status._profiles) == 2
    assert status._profiles["test2"]["count"] == 1
    assert status._profiles["test2"]["max"] == 1
    assert status._profiles["test2"]["min"] == 1
    assert status._profiles["test2"]["total"] == 1
    status.record("test2", 1)
    assert status._profiles["test2"]["count"] == 2
    assert status._profiles["test2"]["max"] == 1
    assert status._profiles["test2"]["min"] == 1
    assert status._profiles["test2"]["total"] == 2
    # test measure
    with status.measure("no-op"):
        pass
    assert len(status._profiles) == 3
    assert "no-op" in status._profiles
    assert len(tuple(status.profile_entries())) == 3


def test_status_08(mocker, tmp_path):
    """test ReadOnlyStatus() - purge expired entries"""
    fake_time = mocker.patch("grizzly.common.status.time", autospec=True)
    db_file = tmp_path / "status.db"
    # purge due to life_time
    fake_time.return_value = 1.0
    status = Status(123, 1.0, db_file, life_time=10)
    status.report(force=True)
    assert any(ReadOnlyStatus.load_all(db_file, time_limit=60))
    fake_time.return_value = 20.0
    Status(456, 20.0, db_file, life_time=10)
    assert not any(ReadOnlyStatus.load_all(db_file, time_limit=60))
    # purge due matching pid
    fake_time.return_value = 1.0
    status = Status(123, 1.0, db_file, life_time=10)
    status.report(force=True)
    assert any(ReadOnlyStatus.load_all(db_file, time_limit=60))
    Status(123, 1.0, db_file, life_time=10)
    assert not any(ReadOnlyStatus.load_all(db_file, time_limit=60))


def test_reduce_status_01(mocker, tmp_path):
    """test ReductionStatus()"""
    mocker.patch("grizzly.common.status.time", autospec=True, return_value=1.0)
    strategies = [f"strategy_{idx}" for idx in range(5)]

    def fake_tc_size():
        return 47

    status = ReductionStatus.start(
        tmp_path / "status.db",
        strategies=strategies,
        testcase_size_cb=fake_tc_size,
    )
    assert status is not None
    assert status.analysis == {}
    assert status.attempts == 0
    assert status.iterations == 0
    assert status.run_params == {}
    assert status.signature_info == {}
    assert status.successes == 0
    assert status.current_strategy_idx is None
    assert status._testcase_size_cb is fake_tc_size
    assert status.crash_id is None
    assert status.finished_steps == []
    assert status._in_progress_steps == []
    assert status.strategies == strategies
    assert status._db_file is not None
    assert status.pid is not None
    assert status.timestamp > 0.0
    assert status._current_size is None


@mark.parametrize(
    "cur_time, force, reported",
    [
        # attempt report before REPORT_RATE elapses
        (count(start=1.0, step=0.1), False, False),
        # force report
        (count(start=1.0, step=0.1), True, True),
        # attempt report after REPORT_RATE elapses
        (count(start=1.0, step=REPORT_RATE + 1.0), False, True),
    ],
)
def test_reduce_status_02(mocker, tmp_path, cur_time, force, reported):
    """test ReductionStatus.report()"""
    mocker.patch("grizzly.common.status.time", autospec=True, side_effect=cur_time)
    status = ReductionStatus.start(tmp_path / "status.db", testcase_size_cb=lambda: 47)
    assert status.report(force=force) == reported


def test_reduce_status_03(tmp_path):
    """test ReductionStatus.load_all()"""
    db_file = tmp_path / "status.db"
    strategies = [f"strategy_{idx}" for idx in range(5)]
    # create simple entry
    status = ReductionStatus.start(
        db_file,
        strategies=strategies,
        testcase_size_cb=lambda: 47,
    )
    loaded = tuple(ReductionStatus.load_all(db_file))
    assert len(loaded) == 1
    loaded = loaded[0]
    assert status.analysis == loaded.analysis
    assert status.attempts == loaded.attempts
    assert status.iterations == loaded.iterations
    assert status.run_params == loaded.run_params
    assert status.signature_info == loaded.signature_info
    assert status.successes == loaded.successes
    assert status.current_strategy_idx == loaded.current_strategy_idx
    assert loaded._testcase_size_cb is None
    assert status.crash_id == loaded.crash_id
    assert status.finished_steps == loaded.finished_steps
    assert status._in_progress_steps == loaded._in_progress_steps
    assert status.strategies == loaded.strategies
    assert status._db_file is not None
    assert status.pid == loaded.pid
    assert status.timestamp == loaded.timestamp
    assert loaded._current_size == 47
    assert loaded._testcase_size() == 47


def test_reduce_status_04(mocker, tmp_path):
    """test ReductionStatus.load_all()"""
    getpid = mocker.patch("grizzly.common.status.getpid", autospec=True)
    db_file = tmp_path / "status.db"
    for pid in range(5):
        getpid.return_value = pid
        ReductionStatus.start(
            db_file,
            testcase_size_cb=lambda: 47,
        )
    assert len(tuple(ReductionStatus.load_all(db_file))) == 5


def test_reduce_status_05(mocker, tmp_path):
    """test ReductionStatus milestone measurements"""
    strategies = [f"strategy_{idx}" for idx in range(5)]

    # (time, testcase_size) steps to manually advance through
    ticks = [
        (0, 1000),
        (1, 900),
        (2, 800),
        (3, 700),
        (4, 600),
        (5, 500),
    ]

    mocker.patch(
        "grizzly.common.status.time",
        autospec=True,
        side_effect=lambda: ticks[0][0],
    )
    testcase_size_cb = mocker.Mock(side_effect=lambda: ticks[0][1])
    status = ReductionStatus.start(
        tmp_path / "status.db",
        strategies=strategies,
        testcase_size_cb=testcase_size_cb,
    )
    status.record("begin")
    assert status.original.name == "begin"
    assert status.total.name == "begin"
    assert status.current_strategy.name == "begin"
    with status.measure("overall"):
        assert status.original.name == "begin"
        assert status.total.name == "overall"
        assert status.current_strategy.name == "overall"
        for idx in range(5):
            with status.measure(strategies[idx]):
                ticks.pop(0)
                status.attempts += 2
                status.successes += 1
                status.iterations += 10
                assert status.original.name == "begin"
                assert status.total.name == "overall"
                assert status.current_strategy.name == f"strategy_{idx}"
    assert status.finished_steps == [
        ReductionStep("begin", None, None, None, 1000, None),
        ReductionStep("strategy_0", 1, 1, 2, 900, 10),
        ReductionStep("strategy_1", 1, 1, 2, 800, 10),
        ReductionStep("strategy_2", 1, 1, 2, 700, 10),
        ReductionStep("strategy_3", 1, 1, 2, 600, 10),
        ReductionStep("strategy_4", 1, 1, 2, 500, 10),
        ReductionStep("overall", 5, 5, 10, 500, 50),
    ]
    assert status.original.name == "begin"
    assert status.total.name == "overall"
    assert status.current_strategy.name == "overall"


def test_reduce_status_06(mocker, tmp_path):
    """test ReductionStatus in-progress milestones"""
    mocker.patch("grizzly.common.status.time", autospec=True, return_value=1.0)
    status = ReductionStatus.start(tmp_path / "status.db", testcase_size_cb=lambda: 47)
    with status.measure("milestone"):
        assert len(status.finished_steps) == 0
        status2 = status.copy()
        assert len(status2.finished_steps) == 1

    assert status.analysis == status2.analysis
    assert status.attempts == status2.attempts
    assert status.iterations == status2.iterations
    assert status.run_params == status2.run_params
    assert status.signature_info == status2.signature_info
    assert status.successes == status2.successes
    assert status.current_strategy_idx == status2.current_strategy_idx
    assert status._testcase_size_cb is status2._testcase_size_cb
    assert status.crash_id == status2.crash_id
    assert status.finished_steps == status2.finished_steps
    assert status._in_progress_steps == status2._in_progress_steps
    assert status.strategies == status2.strategies
    assert status._db_file is not None
    assert status2._db_file is not None
    assert status.pid == status2.pid
    assert status.timestamp == status2.timestamp

    with status.measure("milestone2"):
        status.report(force=True)

        loaded_status = tuple(ReductionStatus.load_all(tmp_path / "status.db"))
        assert len(loaded_status) == 1
        loaded_status = loaded_status[0]

    assert loaded_status.finished_steps == status.finished_steps[:1]
    assert len(loaded_status._in_progress_steps) == 1

    loaded_status = loaded_status.copy()
    assert len(loaded_status.finished_steps) == 2
    assert len(loaded_status._in_progress_steps) == 0
    assert loaded_status.original == status.original
    for field in fields(ReductionStep):
        if field.name == "size":
            continue
        assert getattr(loaded_status.total, field.name) == getattr(
            status.total, field.name
        )
    assert loaded_status.total.size is None


def test_reduce_status_07(mocker, tmp_path):
    """test ReductionStatus metadata"""
    reporter = mocker.Mock(spec_set=FuzzManagerReporter)
    status = ReductionStatus.start(
        tmp_path / "status.db",
        testcase_size_cb=lambda: 47,
        crash_id=123,
    )
    status.analysis["thing"] = "done"
    status.record("init")
    status.run_params["knob"] = "turned"
    status.signature_info["dumb"] = True
    status.add_to_reporter(reporter)
    assert reporter.add_extra_metadata.call_args_list == [
        mocker.call("reducer-stats", [astuple(s) for s in status.finished_steps]),
        mocker.call("reducer-analysis", status.analysis),
        mocker.call("reducer-params", status.run_params),
        mocker.call("reducer-sig", status.signature_info),
        mocker.call("reducer-input", status.crash_id),
    ]


@mark.parametrize(
    "keys, counts, limit",
    [
        # no records
        (["a"], [0], 1),
        # single record (not frequent)
        (["a"], [1], 2),
        # single record (frequent)
        (["a"], [1], 1),
        # single record no limit
        (["a"], [1], 0),
        # multiple records
        (["a", "b", "c"], [1, 2, 10], 5),
    ],
)
def test_report_counter_01(tmp_path, keys, counts, limit):
    """test ResultCounter functionality"""
    db_path = tmp_path / "storage.db"
    counter = ResultCounter(1, db_path, report_limit=limit)
    for report_id, counted in zip(keys, counts):
        assert counter.get(report_id) == ResultEntry(report_id, 0, None)
        assert not counter.is_frequent(report_id)
        # call count() with report_id 'counted' times
        for current in range(1, counted + 1):
            assert counter.count(report_id, "desc") == (current, (current == 1))
        # test get()
        if sum(counts) > 0:
            assert counter.get(report_id) == ResultEntry(report_id, counted, "desc")
        else:
            assert counter.get(report_id) == ResultEntry(report_id, counted, None)
        # test is_frequent()
        if counted > limit > 0:
            assert counter.is_frequent(report_id)
        elif limit > 0:
            assert not counter.is_frequent(report_id)
            # test mark_frequent()
            counter.mark_frequent(report_id)
            assert counter.is_frequent(report_id)
        else:
            assert limit == 0
    for result in counter:
        assert result.count > 0
    assert counter.total == sum(counts)


def test_report_counter_02(mocker, tmp_path):
    """test ResultCounter multi instance functionality"""
    fake_time = mocker.patch("grizzly.common.status.time", autospec=True)
    fake_time.return_value = 1
    db_path = tmp_path / "storage.db"
    counter_a = ResultCounter(1, db_path, report_limit=0)
    counter_b = ResultCounter(2, db_path, report_limit=1)
    counter_c = ResultCounter(3, db_path, report_limit=2)
    # local counts are 0, global (all counters) count is 0
    assert not counter_a.is_frequent("a")
    assert not counter_b.is_frequent("a")
    assert not counter_c.is_frequent("a")
    # local (counter_a, bucket a) count is 1, global (all counters) count is 1
    assert counter_a.count("a", "desc") == (1, True)
    assert not counter_a.is_frequent("a")
    assert not counter_b.is_frequent("a")
    assert not counter_c.is_frequent("a")
    # local (counter_b, bucket a) count is 1, global (all counters) count is 2
    assert counter_b.count("a", "desc") == (1, False)
    assert not counter_a.is_frequent("a")
    assert not counter_b.is_frequent("a")
    assert not counter_c.is_frequent("a")
    # local (counter_b, bucket a) count is 2, global (all counters) count is 3
    # locally exceeded
    assert counter_b.count("a", "desc") == (2, False)
    assert counter_b.is_frequent("a")
    # local (counter_c, bucket a) count is 1, global (all counters) count is 4
    assert counter_c.count("a", "desc") == (1, False)
    assert not counter_a.is_frequent("a")
    assert counter_b.is_frequent("a")
    assert not counter_c.is_frequent("a")
    # local (counter_a, bucket a) count is 2, global (all counters) count is 5
    # no limit
    assert counter_a.count("a", "desc") == (2, False)
    assert not counter_a.is_frequent("a")
    # local (counter_c, bucket a) count is 2, global (all counters) count is 6
    # locally not exceeded, globally exceeded
    assert counter_c.count("a", "desc") == (2, False)
    assert counter_c.is_frequent("a")
    # local (counter_a, bucket x) count is 0, global (all counters) count is 0
    assert not counter_a.is_frequent("x")
    # remove 'expired' reports
    fake_time.return_value = 1000
    counter_d = ResultCounter(4, db_path, life_time=10, report_limit=2)
    # local (counter_d, bucket a) count is 0, global (all counters) count is 0
    assert not counter_d.is_frequent("a")
    assert counter_a.total == 2
    assert counter_b.total == 2
    assert counter_c.total == 2
    assert counter_d.total == 0


def test_report_counter_03(mocker, tmp_path):
    """test ReadOnlyResultCounter.load()"""
    fake_time = mocker.patch("grizzly.common.status.time", autospec=True)
    fake_time.return_value = 1
    db_path = tmp_path / "storage.db"
    # load empty db
    assert not ReadOnlyResultCounter.load(db_path, 10)
    # create counter
    counter = ResultCounter(123, db_path, life_time=1)
    counter.count("a", "desc_a")
    fake_time.return_value = 2
    counter.count("a", "desc_a")
    fake_time.return_value = 3
    counter.count("b", "desc_b")
    # filter out reports by time
    fake_time.return_value = 4
    # last 1 second
    assert not ReadOnlyResultCounter.load(db_path, 1)
    # last 2 seconds
    loaded = ReadOnlyResultCounter.load(db_path, 2)[0]
    assert loaded.total == 1
    assert loaded.get("b") == ResultEntry("b", 1, "desc_b")
    # last 3 seconds
    loaded = ReadOnlyResultCounter.load(db_path, 3)[0]
    assert loaded.get("a") == ResultEntry("a", 2, "desc_a")
    assert loaded.total == 3
    # increase time limit
    fake_time.return_value = 4
    loaded = ReadOnlyResultCounter.load(db_path, 10)[0]
    assert loaded.total == counter.total == 3
    assert loaded.get("a") == ResultEntry("a", 2, "desc_a")
    assert loaded.get("b") == ResultEntry("b", 1, "desc_b")


def test_report_counter_04(mocker, tmp_path):
    """test ResultCounter remove expired entries"""
    fake_time = mocker.patch("grizzly.common.status.time", autospec=True)
    fake_time.return_value = 1
    db_path = tmp_path / "storage.db"
    counter = ResultCounter(123, db_path, life_time=0)
    counter.count("a", "desc_a")
    fake_time.return_value = 100
    counter.count("b", "desc_b")
    loaded = ReadOnlyResultCounter.load(db_path, time_limit=100)[0]
    assert loaded.total == 2
    # set life_time to zero to skip removing expired results
    ResultCounter(124, db_path, life_time=0)
    loaded = ReadOnlyResultCounter.load(db_path, time_limit=100)[0]
    assert loaded.total == 2
    # clear expired records from database by setting life_time
    ResultCounter(125, db_path, life_time=10)
    loaded = ReadOnlyResultCounter.load(db_path, time_limit=100)[0]
    assert loaded.total == 1
    # clear expired records from database by using duplicate pid
    ResultCounter(123, db_path, life_time=1000)
    assert not ReadOnlyResultCounter.load(db_path, time_limit=100)


@mark.parametrize(
    "buckets, ratio, iterations, blockers",
    [
        # no results
        ([], 1, 1, 0),
        # one result seen once (not blocker since count == 1)
        ([("uid1", "sig1", 1)], 1, 1, 0),
        # one result seen 10x (not blocker)
        ([("uid1", "sig1", 10)], 100, 10000, 0),
        # one result seen 10x (blocker)
        ([("uid1", "sig1", 10)], 100, 1000, 1),
        # one result seen 95x (blocker)
        ([("uid1", "sig1", 95)], 100, 1000, 1),
        # multiple results seen once (not blocker since count == 1)
        ([("uid1", "sig1", 1), ("uid2", "sig2", 1)], 1, 1, 0),
        # multiple results seen once (one blockers)
        ([("uid1", "sig1", 1), ("uid2", "sig2", 10)], 1000, 100, 1),
        # multiple results seen once (two blockers)
        ([("uid1", "sig1", 99), ("uid2", "sig2", 10)], 1000, 100, 2),
    ],
)
def test_simple_report_counter_01(buckets, ratio, iterations, blockers):
    """test SimpleResultCounter.blockers()"""
    counter = SimpleResultCounter(123)
    # populate counter
    for report_id, desc, total in buckets:
        counter._desc[report_id] = desc
        counter._count[report_id] = total
    # check for blockers
    assert len(tuple(counter.blockers(iterations, iters_per_result=ratio))) == blockers


def test_db_version_check_01(tmp_path):
    """test _db_version_check()"""
    db_path = tmp_path / "storage.db"
    with closing(connect(db_path, timeout=10)) as con:
        # empty db
        assert _db_version_check(con, expected=DB_VERSION)
        # no update needed
        assert not _db_version_check(con, expected=DB_VERSION)
        # add db contents
        Status.start(db_path)
        # force update
        assert _db_version_check(con, expected=DB_VERSION + 1)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from json import loads
from logging import getLogger
from typing import TYPE_CHECKING

from ..common.frontend import configure_logging
from ..common.fuzzmanager import Bucket, CrashEntry, load_fm_data
from .args import ReplayFuzzManagerIDArgs
from .replay import ReplayManager

if TYPE_CHECKING:
    from argparse import Namespace

LOG = getLogger(__name__)


def main(args: Namespace | None = None) -> int:
    """CLI for `grizzly.replay.crash`.

    Arguments:
        args: Result from `ReplayFuzzManagerIDArgs.parse_args`.

    Returns:
        Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
    """
    args = args or ReplayFuzzManagerIDArgs().parse_args()
    configure_logging(args.log_level)
    with load_fm_data(args.input, load_bucket=not args.sig) as (crash, bucket):
        LOG.info("Loaded crash %d from FuzzManager", crash.crash_id)
        # call grizzly.replay
        return ReplayManager.main(modify_args(args, crash, bucket))


def modify_args(args: Namespace, crash: CrashEntry, bucket: Bucket | None) -> Namespace:
    """

    Arguments:
        args: Result from `ReplayFuzzManagerIDArgs.parse_args`.
        crash: Crash entry to process.
        bucket: Bucket that contains crash.

    Returns:
        Modified arguments.
    """
    # if we did not pass --sig, and the crash is not bucketed
    # auto-generate a signature from crash data so we know what to expect
    if not args.sig and not bucket:
        try:
            fm_sig_file = crash.create_signature(args.binary)
            meta = loads(fm_sig_file.with_suffix(".metadata").read_text())
            LOG.info(
                "Using crash data to generate signature: %s",
                meta["shortDescription"],
            )
            args.sig = fm_sig_file
        except RuntimeError as exc:
            LOG.warning("Failed to generate signature from crash data: %s", exc)

    args.original_crash_id = args.input
    # use the newest test case when not using a harness and test_index is not specified
    if args.no_harness and not args.test_index:
        args.test_index = [-1]
    args.input = crash.testcases(subset=args.test_index, ext=args.test_ext)
    # set tool name using crash entry
    if args.tool is None:
        LOG.info("Setting default --tool=%s from CrashEntry", crash.tool)
        args.tool = crash.tool
    # load signature if needed
    if bucket is not None:
        args.sig = bucket.signature_path()
        LOG.info("Using signature from bucket: %d", bucket.bucket_id)

    return args


if __name__ == "__main__":
    raise SystemExit(main())



"""
Sapphire unit tests
"""

# pylint: disable=protected-access

import socket
from hashlib import sha1
from itertools import count, repeat
from os import urandom
from pathlib import Path
from platform import system
from random import choices, getrandbits
from threading import Lock
from urllib.parse import quote, urlsplit

from pytest import mark, raises

from .certificate_bundle import CertificateBundle
from .core import Sapphire, create_listening_socket
from .job import Served
from .server_map import ServerMap
from .worker import Worker


class _TestFile:
    def __init__(self, url, url_prefix=None):
        assert isinstance(url, str)
        self.code = None
        self.content_type = None
        self.custom_request = None
        self.file = f"{url_prefix}{url}" if url_prefix else url
        self.len_org = 0  # original file length
        self.len_srv = 0  # served file length
        self.lock = Lock()
        self.hash_org = None
        self.hash_srv = None
        self.requested = 0  # number of times file was requested
        url = urlsplit(self.file.replace("\\", "/"))
        self.url = (
            "?".join((quote(url.path), url.query)) if url.query else quote(url.path)
        )

    @classmethod
    def create(cls, fname, path, data=b"Test!", calc_hash=False, url_prefix=None):
        test = cls(fname, url_prefix=url_prefix)
        with (path / fname).open("w+b") as out_fp:
            out_fp.write(data)
            test.len_org = out_fp.tell()
            if calc_hash:
                out_fp.seek(0)
                test.hash_org = sha1(out_fp.read()).hexdigest()
        return test


@mark.parametrize("files", [1, 100])
def test_sapphire_01(client, tmp_path, files):
    """test serving files"""
    _TestFile.create("unrelated.bin", tmp_path)
    to_serve = [
        _TestFile.create(
            f"test_{i:04d}.html", tmp_path, data=urandom(5), calc_hash=True
        )
        for i in range(files)
    ]
    # all files are required
    required = [x.file for x in to_serve]
    with Sapphire(timeout=30) as serv:
        assert serv.timeout == 30
        assert serv.scheme == "http"
        client.launch("127.0.0.1", serv.port, to_serve)
        status, served = serv.serve_path(tmp_path, required_files=required)
    assert status == Served.ALL
    assert "unrelated.bin" not in served
    assert len(required) == len(served) == len(to_serve)
    assert client.wait(timeout=10)
    for t_file in to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org
        assert t_file.url in served


@mark.parametrize(
    "files, req_idx",
    [
        # multiple files (skip optional)
        (5, 0),
        # multiple files (serve optional)
        (5, 4),
    ],
)
def test_sapphire_02(client, tmp_path, files, req_idx):
    """test serving files"""
    _TestFile.create("unrelated.bin", tmp_path)
    to_serve = [
        _TestFile.create(
            f"test_{i:04d}.html", tmp_path, data=urandom(5), calc_hash=True
        )
        for i in range(files)
    ]
    required = to_serve[req_idx].file
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve, in_order=True)
        status, served = serv.serve_path(tmp_path, required_files=[required])
    assert status == Served.ALL
    assert "unrelated.bin" not in served
    assert required in served
    assert len(served) >= (req_idx + 1)
    assert client.wait(timeout=10)
    for t_file in to_serve:
        if t_file.file in served:
            assert t_file.code == 200
            assert t_file.len_srv == t_file.len_org


def test_sapphire_03(client, tmp_path):
    """test requesting invalid files (404 and 403)"""
    root_dir = tmp_path / "root"
    root_dir.mkdir()
    invalid = Path(__file__)
    to_serve = [
        # 0 - missing file
        _TestFile("does_not_exist.html"),
        # 1 - add invalid file
        _TestFile(str(invalid.resolve())),
        # 2 - add file in parent of root_dir
        _TestFile.create(
            "no_access.html", tmp_path, data=b"no_access", url_prefix="../"
        ),
        # 3 - add valid test
        _TestFile.create("test_case.html", root_dir),
    ]
    required = [to_serve[-1].file]
    assert (tmp_path / "no_access.html").is_file()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve, in_order=True)
        status, served = serv.serve_path(root_dir, required_files=required)
    assert status == Served.ALL
    assert len(served) == 1
    assert client.wait(timeout=10)
    assert to_serve[0].code == 404
    assert invalid.name in to_serve[1].file
    assert to_serve[1].code == 404
    assert "no_access.html" in to_serve[2].file
    assert to_serve[2].code == 404
    assert to_serve[3].code == 200


def test_sapphire_04(tmp_path):
    """test timeout of the server"""
    with Sapphire(timeout=0.01) as serv:
        assert serv.timeout == 0.01
        to_serve = [_TestFile.create("test_case.html", tmp_path)]
        status, served = serv.serve_path(tmp_path, required_files=[to_serve[0].file])
    assert status == Served.TIMEOUT
    assert not served


def test_sapphire_05(client, tmp_path):
    """test only serving some files (Served.REQUEST)"""
    cb_status = {"count": 0}

    def is_running():
        cb_status["count"] += 1
        return cb_status["count"] < 3  # return false after 2nd call

    to_serve = [_TestFile.create(f"test_{i}.html", tmp_path) for i in range(3)]
    with Sapphire() as serv:
        client.launch("127.0.0.1", serv.port, to_serve[1:])
        status, served = serv.serve_path(
            tmp_path, continue_cb=is_running, required_files=[x.file for x in to_serve]
        )
    assert status == Served.REQUEST
    assert len(served) < len(to_serve)


def test_sapphire_06(client, tmp_path):
    """test serving interesting sized files"""
    tests = [
        {"size": Worker.DEFAULT_TX_SIZE, "name": "even.html"},
        {"size": Worker.DEFAULT_TX_SIZE - 1, "name": "minus_one.html"},
        {"size": Worker.DEFAULT_TX_SIZE + 1, "name": "plus_one.html"},
        {"size": Worker.DEFAULT_TX_SIZE * 2, "name": "double.html"},
        {"size": 1, "name": "one.html"},
        {"size": 0, "name": "zero.html"},
    ]
    for test in tests:
        test["file"] = _TestFile.create(
            test["name"],
            tmp_path,
            data="".join(choices("ABCD1234", k=test["size"])).encode("ascii"),
            calc_hash=True,
        )
    required = [test["file"].file for test in tests]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test["file"] for test in tests])
        status, served_list = serv.serve_path(tmp_path, required_files=required)
    assert status == Served.ALL
    assert len(served_list) == len(tests)
    assert client.wait(timeout=10)
    for test in tests:
        assert test["file"].code == 200
        assert test["file"].len_srv == test["size"]
        assert test["file"].hash_srv == test["file"].hash_org


def test_sapphire_07(client, tmp_path):
    """test serving a large (100MB) file"""
    t_file = _TestFile("test_case.html")
    data_hash = sha1()
    with (tmp_path / t_file.file).open("wb") as test_fp:
        # write 100MB of 'A'
        data = b"A" * (100 * 1024)  # 100KB of 'A'
        for _ in range(1024):
            test_fp.write(data)
            data_hash.update(data)
    t_file.hash_org = data_hash.hexdigest()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [t_file])
        assert serv.serve_path(tmp_path, required_files=[t_file.file])[0] == Served.ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == (100 * 1024 * 1024)
    assert t_file.hash_srv == t_file.hash_org


def test_sapphire_08(client, tmp_path):
    """test serving a binary file"""
    t_file = _TestFile.create("test.html", tmp_path, data=urandom(512), calc_hash=True)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [t_file])
        assert serv.serve_path(tmp_path, required_files=[t_file.file])[0] == Served.ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == t_file.len_org
    assert t_file.hash_srv == t_file.hash_org


def test_sapphire_09():
    """test requested port is used"""
    test_port = 0x1337
    with Sapphire(port=test_port, timeout=1) as serv:
        assert test_port == serv.port


def test_sapphire_10(client, tmp_path):
    """test serving multiple content types"""
    to_serve = [
        _TestFile.create("test_case.html", tmp_path),
        # create binary file without an extension
        _TestFile.create("test_case", tmp_path, data=urandom(5)),
    ]
    required = [x.file for x in to_serve]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve)
        assert serv.serve_path(tmp_path, required_files=required)[0] == Served.ALL
    assert client.wait(timeout=10)
    content_types = set()
    for test in to_serve:
        assert test.code == 200
        assert test.len_srv == test.len_org
        if test.file.endswith(".html"):
            content_type = "text/html"
        else:
            content_type = "application/octet-stream"
        content_types.add(content_type)
        assert test.content_type == content_type
    assert len(content_types) == 2


def test_sapphire_11(client, tmp_path):
    """test calling serve_path multiple times"""
    with Sapphire(timeout=10) as serv:
        for i in range(3):
            name = f"test_{i}.html"
            test = _TestFile.create(name, tmp_path)
            client.launch("127.0.0.1", serv.port, [test])
            assert serv.serve_path(tmp_path, required_files=[name])[0] == Served.ALL
            assert client.wait(timeout=10)
            client.close()
            assert test.code == 200
            assert test.len_srv == test.len_org
            (tmp_path / test.file).unlink()


def test_sapphire_12(client, tmp_path):
    """test non required mapped redirects"""
    smap = ServerMap()
    smap.set_redirect("test_url", "blah", required=False)
    test = _TestFile.create("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test])
        assert (
            serv.serve_path(tmp_path, server_map=smap, required_files=[test.file])[0]
            == Served.ALL
        )
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org


@mark.parametrize(
    "path, query",
    [
        # simple path
        ("test.html", None),
        # simple path with query
        ("test.html", "foo=1&bar=2"),
        # non-alphanumeric chars (valid characters to use on filesystem)
        ("!@$%^&(_+-=[]),'~`{}", None),
        # extended ascii chars
        ("€d’é-ñÿ", None),  # noqa: RUF001
    ],
)
def test_sapphire_13(client, tmp_path, path, query):
    """test required mapped redirects"""
    smap = ServerMap()
    with Sapphire(timeout=10) as serv:
        # target will be requested indirectly via the redirect
        target = _TestFile.create(path, tmp_path, data=b"Redirect DATA!")
        request_path = "redirect" if query is None else f"redirect?{query}"
        redirect = _TestFile(request_path)
        # point "redirect" at target
        smap.set_redirect("redirect", target.file, required=True)
        client.launch("127.0.0.1", serv.port, [redirect])
        status, served = serv.serve_path(
            tmp_path, server_map=smap, required_files=[target.file]
        )
    assert status == Served.ALL
    assert len(served) == 1
    assert client.wait(timeout=10)
    assert redirect.code == 200
    assert redirect.len_srv == target.len_org


def test_sapphire_14(client, tmp_path):
    """test include directories and permissions"""
    inc1_path = tmp_path / "inc1"
    inc2_path = tmp_path / "inc2"
    root_path = tmp_path / "root"
    inc1_path.mkdir()
    inc2_path.mkdir()
    root_path.mkdir()
    to_serve = []
    smap = ServerMap()
    with Sapphire(timeout=10) as serv:
        # add files to inc dirs
        inc1 = _TestFile.create("included_file1.html", inc1_path, data=b"blah....1")
        to_serve.append(inc1)
        # add a nested dir
        nest_path = inc1_path / "nested"
        nest_path.mkdir()
        # add file in a nested dir in inc1
        nest = _TestFile.create(
            "nested_file.html", nest_path, data=b"blah... .nested", url_prefix="nested/"
        )
        assert nest_path / "nested_file.html"
        to_serve.append(nest)
        # test 404 in nested dir in inc1
        nest_404 = _TestFile("nested/nested_file_404.html")
        to_serve.append(nest_404)
        # test path mounted somewhere other than /
        inc2 = _TestFile.create(
            "included_file2.html", inc2_path, data=b"blah....2", url_prefix="inc_test/"
        )
        to_serve.append(inc2)
        # test 404 in include dir
        inc404 = _TestFile("inc_test/included_file_404.html")
        assert not (nest_path / "included_file_404.html").is_file()
        to_serve.append(inc404)
        # test 404 with file outside of include path
        inc_ext = _TestFile.create(
            "no_access.html", tmp_path, data=b"no_access", url_prefix="inc_test/../"
        )
        assert (tmp_path / "no_access.html").is_file()
        to_serve.append(inc_ext)
        # test file (used to keep sever job alive)
        test = _TestFile.create("test_case.html", root_path)
        to_serve.append(test)
        # add include paths
        smap.set_include("/", inc1_path)  # mount at '/'
        smap.set_include("inc_test", inc2_path)  # mount at '/inc_test'
        client.launch("127.0.0.1", serv.port, to_serve, in_order=True)
        status, served = serv.serve_path(
            root_path, server_map=smap, required_files=[x.file for x in to_serve]
        )
    assert status == Served.ALL
    assert "test_case.html" in served
    assert "included_file1.html" in served
    assert served["included_file1.html"] == (inc1_path / "included_file1.html")
    assert "inc_test/included_file2.html" in served
    assert served["inc_test/included_file2.html"] == (inc2_path / "included_file2.html")
    assert "nested/nested_file.html" in served
    assert served["nested/nested_file.html"] == (nest_path / "nested_file.html")
    assert client.wait(timeout=10)
    assert inc1.code == 200
    assert inc2.code == 200
    assert nest.code == 200
    assert test.code == 200
    assert nest_404.code == 404
    assert inc404.code == 404
    assert inc_ext.code == 404


@mark.parametrize(
    "query, required",
    [
        # dynamic response not required
        (None, False),
        # dynamic response is required
        (None, True),
        # dynamic response with query string
        ("test123", True),
        # dynamic response with empty query string
        ("", True),
    ],
)
def test_sapphire_15(client, tmp_path, query, required):
    """test dynamic response"""
    _data = b"dynamic response -- TEST DATA!"
    # build request
    path = "dyn_test"
    request = path if query is None else f"{path}?{query}"

    # setup custom callback
    def dr_callback(data):
        if query:
            assert data == query
        else:
            assert data == ""
        return _data

    smap = ServerMap()
    smap.set_dynamic_response(
        path, dr_callback, mime_type="text/plain", required=required
    )
    # create files
    test_dr = _TestFile(request)
    test_dr.len_org = len(_data)
    test_dr.hash_org = sha1(_data).hexdigest()
    test = _TestFile.create("test_case.html", tmp_path)
    if required:
        req_files = []
        files = [test, test_dr]
    else:
        req_files = [test.file]
        files = [test_dr, test]
    # test request
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files, in_order=True)
        assert (
            serv.serve_path(tmp_path, required_files=req_files, server_map=smap)[0]
            == Served.ALL
        )
    assert client.wait(timeout=10)
    if not required:
        assert test.code == 200
        assert test.len_srv == test.len_org
    assert test_dr.code == 200
    assert test_dr.len_srv == test_dr.len_org
    assert test_dr.hash_srv == test_dr.hash_org


def test_sapphire_16(client_factory, tmp_path):
    """test pending_files == 0 in worker thread"""
    client_defer = client_factory(rx_size=2)
    # server should shutdown while this file is being served
    test_defer = _TestFile.create("defer_test.html", tmp_path)
    test = _TestFile.create("test_case.html", tmp_path, data=b"112233")
    with Sapphire(timeout=10) as serv:
        # this test needs to wait just long enough to have the required file served
        # but not too long or the connection will be closed by the server
        client_defer.launch(
            "127.0.0.1", serv.port, [test_defer], delay=0.1, indicate_failure=True
        )
        client = client_factory(rx_size=2)
        client.launch("127.0.0.1", serv.port, [test], throttle=0.1)
        assert serv.serve_path(tmp_path, required_files=[test.file])[0] == Served.ALL
    assert client_defer.wait(timeout=10)
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test_defer.code == 0


def test_sapphire_17(client, tmp_path):
    """test handling an invalid request"""
    bad_test = _TestFile("bad.html")
    bad_test.custom_request = b"a bad request...0+%\xef\xb7\xba\r\n"
    test = _TestFile.create("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [bad_test, test], in_order=True)
        assert serv.serve_path(tmp_path, required_files=[test.file])[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert bad_test.code == 400


def test_sapphire_18(client, tmp_path):
    """test handling an empty request"""
    bad_test = _TestFile("bad.html")
    bad_test.custom_request = b""
    test = _TestFile.create("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch(
            "127.0.0.1",
            serv.port,
            [bad_test, test],
            indicate_failure=True,
            in_order=True,
        )
        assert serv.serve_path(tmp_path, required_files=[test.file])[0] == Served.ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert bad_test.code == 0


def test_sapphire_19(client_factory, tmp_path):
    """test requesting multiple files via multiple connections"""
    to_serve = [
        _TestFile.create(f"test_{i:03d}.html", tmp_path, data=b"AAAA") for i in range(2)
    ]
    max_workers = 20
    with Sapphire(max_workers=max_workers, timeout=60) as serv:
        clients = []
        try:
            clients = [client_factory(rx_size=1) for _ in range(max_workers)]
            for client in clients:
                client.launch(
                    "127.0.0.1", serv.port, to_serve, in_order=True, throttle=0.05
                )
            required = [x.file for x in to_serve]
            status, served = serv.serve_path(tmp_path, required_files=required)
            # call serv.close() instead of waiting for the clients to timeout
            serv.close()
        finally:
            for client in clients:
                assert client.wait(timeout=10)
                client.close()
    assert status == Served.ALL
    assert len(to_serve) == len(served)
    for t_file in to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org


def test_sapphire_20(client_factory, tmp_path):
    """test all request types via multiple connections"""

    def _dyn_test_cb(_):
        return b"A" if getrandbits(1) else b"AA"

    smap = ServerMap()
    with Sapphire(max_workers=10, timeout=60) as serv:
        to_serve = []
        required = []
        for i in range(50):
            # add required files
            to_serve.append(
                _TestFile.create(
                    f"test_{i:03d}.html", tmp_path, data=b"A" * ((i % 2) + 1)
                )
            )
            required.append(to_serve[-1].file)
            # add a missing files
            to_serve.append(_TestFile(f"missing_{i:03d}.html"))
            # add optional files
            opt_path = tmp_path / f"opt_{i:03d}.html"
            opt_path.write_bytes(b"A" * ((i % 2) + 1))
            to_serve.append(_TestFile(opt_path.name))
            # add redirects
            redir_target = _TestFile.create(f"redir_{i:03d}.html", tmp_path, data=b"AA")
            to_serve.append(_TestFile(f"redir_{i:03d}"))
            smap.set_redirect(
                to_serve[-1].file, redir_target.file, required=getrandbits(1) > 0
            )
            # add dynamic responses
            to_serve.append(_TestFile(f"dynm_{i:03d}"))
            smap.set_dynamic_response(
                to_serve[-1].file, _dyn_test_cb, mime_type="text/plain"
            )
        clients = []
        for _ in range(100):  # number of clients to spawn
            clients.append(client_factory(rx_size=1))
            throttle = 0.05 if getrandbits(1) else 0
            clients[-1].launch("127.0.0.1", serv.port, to_serve, throttle=throttle)
        assert (
            serv.serve_path(tmp_path, server_map=smap, required_files=required)[0]
            == Served.ALL
        )


def test_sapphire_21(client, tmp_path):
    """test dynamic response with bad callbacks"""
    test_dr = _TestFile("dynm_test")
    smap = ServerMap()
    smap.set_dynamic_response("dynm_test", lambda _: None, mime_type="text/plain")
    test = _TestFile.create("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test_dr, test], in_order=True)
        with raises(TypeError, match="dynamic request callback must return 'bytes'"):
            serv.serve_path(tmp_path, server_map=smap, required_files=[test.file])


def test_sapphire_22(client, tmp_path):
    """test serving to a slow client"""
    t_data = "".join(choices("ABCD1234", k=0x19000)).encode("ascii")  # 100KB
    t_file = _TestFile.create("test_case.html", tmp_path, data=t_data, calc_hash=True)
    # rx_size 10KB and throttle to 0.25 sec, which will be ~50KB/s
    # also taking 2.5 seconds to complete will hopefully find problems
    # with any assumptions that were made
    client.rx_size = 0x2800
    with Sapphire(timeout=60) as serv:
        client.launch("127.0.0.1", serv.port, [t_file], throttle=0.25)
        assert serv.serve_path(tmp_path, required_files=[t_file.file])[0] == Served.ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == t_file.len_org
    assert t_file.hash_srv == t_file.hash_org


def test_sapphire_23(client, tmp_path):
    """test timeout while requesting multiple files"""
    t_data = "".join(choices("ABCD1234", k=1024)).encode("ascii")
    to_serve = [
        _TestFile.create(f"test_{i:03d}.html", tmp_path, data=t_data) for i in range(50)
    ]
    required = [x.file for x in to_serve]
    client.rx_size = 512
    with Sapphire(timeout=1) as serv:  # minimum timeout is 1 second
        client.launch(
            "127.0.0.1", serv.port, to_serve, indicate_failure=True, throttle=0.1
        )
        status, served = serv.serve_path(tmp_path, required_files=required)
    assert status == Served.TIMEOUT
    assert len(served) < len(to_serve)


def test_sapphire_24(client_factory, tmp_path):
    """test Sapphire.serve_path() with forever=True"""
    with Sapphire(timeout=10) as serv:
        test = _TestFile.create("test_case.html", tmp_path)
        clients = [client_factory() for _ in range(3)]
        for client in clients:
            client.launch("127.0.0.1", serv.port, [test], skip_served=False)

        def _test_callback():
            with test.lock:
                return test.requested < 3

        assert (
            serv.serve_path(tmp_path, continue_cb=_test_callback, forever=True)[0]
            == Served.ALL
        )
    for client in clients:
        assert client.wait(timeout=10)
        client.close()
    assert test.requested == 3
    assert test.code == 200
    assert test.len_srv == test.len_org


@mark.parametrize(
    "file_name",
    [
        # space in file name
        "test case.html",
        # non-alphanumeric chars (valid characters to use on filesystem)
        "!@$%^&(_+-=[]),'~`{}",
        # extended ascii chars
        "€d’é-ñÿ",  # noqa: RUF001
    ],
)
def test_sapphire_25(client, tmp_path, file_name):
    """test interesting file names"""
    to_serve = [_TestFile.create(file_name, tmp_path)]
    required = [x.file for x in to_serve]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve)
        assert serv.serve_path(tmp_path, required_files=required)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert all(t_file.code == 200 for t_file in to_serve)


def test_sapphire_26(client, tmp_path):
    """test interesting path string"""
    path = "".join(chr(i) for i in range(256))
    to_serve = [
        # should not trigger crash
        _TestFile(path),
        # used to keep server running
        _TestFile.create("a.html", tmp_path),
    ]
    required = [to_serve[-1].file]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve, in_order=True)
        assert serv.serve_path(tmp_path, required_files=required)[0] == Served.ALL
    assert client.wait(timeout=10)
    assert all(t_file.code is not None for t_file in to_serve)


def test_sapphire_27(mocker):
    """test Sapphire.clear_backlog()"""
    mocker.patch("sapphire.core.perf_counter", autospec=True, side_effect=count())
    mocker.patch("sapphire.core.sleep", autospec=True)
    # test clearing backlog
    pending = mocker.Mock(spec_set=socket.socket)
    pending.accept.side_effect = ((pending, None), OSError, BlockingIOError)
    pending.getblocking.return_value = False
    pending.getsockname.return_value = (None, 1337)
    mocker.patch("sapphire.core.socket", return_value=pending)
    with Sapphire(timeout=10) as serv:
        assert serv.clear_backlog()
        assert serv._socket.accept.call_count == 3
        assert pending.close.call_count == 1
    pending.reset_mock()
    # test hang
    pending.accept.side_effect = None
    pending.accept.return_value = (pending, None)
    with Sapphire(timeout=1) as serv:
        assert not serv.clear_backlog()


@mark.skipif(system() != "Windows", reason="Only supported on Windows")
def test_sapphire_28(client, tmp_path):
    """test serving from path using Windows short file name"""
    wwwroot = tmp_path / "long_path_name_that_can_be_truncated_on_windows"
    wwwroot.mkdir()
    with Sapphire(timeout=10) as serv:
        assert serv.timeout == 10
        test = _TestFile.create("test_case.html", wwwroot)
        client.launch("127.0.0.1", serv.port, [test])
        assert (
            serv.serve_path(tmp_path / "LONG_P~1", required_files=[test.file])[0]
            == Served.ALL
        )
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org


def test_sapphire_29():
    """test Sapphire with certificates"""
    certs = CertificateBundle.create()
    try:
        with Sapphire(timeout=10, certs=certs) as serv:
            assert serv.scheme == "https"
    finally:
        certs.cleanup()


@mark.parametrize(
    "bind,",
    [
        # success
        (None,),
        # failure and success on retry
        (PermissionError("foo", 10013), None),
    ],
)
def test_create_listening_socket_01(mocker, bind):
    """test create_listening_socket()"""
    fake_sleep = mocker.patch("sapphire.core.sleep", autospec=True)
    fake_sock = mocker.patch("sapphire.core.socket", autospec=True)
    fake_sock.return_value.bind.side_effect = bind
    bind_calls = len(bind)
    assert create_listening_socket()
    assert fake_sock.return_value.close.call_count == bind_calls - 1
    assert fake_sock.return_value.setsockopt.call_count == bind_calls
    assert fake_sock.return_value.bind.call_count == bind_calls
    assert fake_sock.return_value.listen.call_count == 1
    assert fake_sleep.call_count == bind_calls - 1


@mark.parametrize(
    "bind, attempts, raised",
    [
        # failure to bind (no retry)
        ((OSError("foo"),), 1, OSError),
        # failure and fail on retry
        (repeat(PermissionError("foo", 10013), 2), 2, PermissionError),
    ],
)
def test_create_listening_socket_02(mocker, bind, attempts, raised):
    """test create_listening_socket() - bind/listen failure"""
    mocker.patch("sapphire.core.sleep", autospec=True)
    fake_sock = mocker.patch("sapphire.core.socket", autospec=True)
    fake_sock.return_value.bind.side_effect = bind
    with raises(raised, match="foo"):
        create_listening_socket(attempts=attempts)
    assert fake_sock.return_value.close.call_count == attempts


def test_create_listening_socket_03(mocker):
    """test create_listening_socket() - fail to find port"""
    fake_sock = mocker.patch("sapphire.core.socket", autospec=True)
    # specify blocked port
    with raises(ValueError, match="Cannot bind to blocked ports"):
        create_listening_socket(port=6000, attempts=1)
    # specify reserved port
    with raises(ValueError, match="Cannot bind to blocked ports"):
        create_listening_socket(port=123, attempts=1)
    # always choose a blocked port
    fake_sock.return_value.getsockname.return_value = (None, 6665)
    with raises(RuntimeError, match="Could not find available port"):
        create_listening_socket(attempts=1)
    assert fake_sock.return_value.listen.call_count == 1
    assert fake_sock.return_value.close.call_count == 1


def test_main_01(mocker, tmp_path):
    """test Sapphire.main()"""
    args = mocker.Mock(path=tmp_path, port=4536, remote=False, timeout=0)
    fake_srv = mocker.patch("sapphire.core.Sapphire.serve_path", autospec=True)
    fake_srv.return_value = (Served.ALL, None)
    Sapphire.main(args)
    fake_srv.return_value = (Served.NONE, None)
    Sapphire.main(args)
    fake_srv.side_effect = KeyboardInterrupt
    Sapphire.main(args)



#!/usr/bin/env python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
stack_hasher will take input text, parse out the stack trace and hash based on
the stack. When using to bucket crashes it is helpful to make two hashes, one
of the entire stack and the second that is less sensitive using the first few
entries on the top of the stack with the offsets removed. This returns a unique
crash id (1st hash) and a bug id (2nd hash). This is not perfect but works very
well in most cases.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import suppress
from hashlib import sha1
from logging import DEBUG, INFO, basicConfig, getLogger
from os.path import basename
from re import compile as re_compile
from re import match as re_match

__all__ = ("Stack",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

# These entries pad out the stack and make bucketing more difficult
IGNORED_FRAMES = (
    "AnnotateMozCrashReason",
    "core::panicking::",
    "mozglue_static::panic_hook",
    "rust_begin_unwind",
    "RustMozCrash",
    "std::panicking::",
    "std::sys_common::backtrace::",
)
LOG = getLogger(__name__)
MAJOR_DEPTH = 5
_RE_FUNC_NAME = re_compile(r"(?P<func>.+?)[\(|\s|\<]{1}")


class StackFrame(ABC):
    __slots__ = ("function", "location", "offset", "stack_line")

    def __init__(
        self,
        function: str | None = None,
        location: str | None = None,
        offset: str | None = None,
        stack_line: str | None = None,
    ) -> None:
        self.function = function
        self.location = location
        self.offset = offset
        self.stack_line = stack_line

    def __str__(self) -> str:
        out = []
        if self.stack_line is not None:
            out.append(f"{int(self.stack_line):02d}")
        if self.function is not None:
            out.append(f"function: {self.function!r}")
        if self.location is not None:
            out.append(f"location: {self.location!r}")
        if self.offset is not None:
            out.append(f"offset: {self.offset!r}")
        return " - ".join(out)

    @classmethod
    @abstractmethod
    def from_line(cls, input_line: str) -> StackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            StackFrame
        """


class MinidumpStackFrame(StackFrame):
    @classmethod
    def from_line(cls, input_line: str) -> MinidumpStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            MinidumpStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        try:
            (
                tid,
                stack_line,
                lib_name,
                func_name,
                file_name,
                line_no,
                offset,
            ) = input_line.split("|")
            # check tid and stack_line are valid
            _ = int(tid)
            _ = int(stack_line)
        except ValueError:
            return None
        sframe = cls(stack_line=stack_line)
        if func_name:
            sframe.function = func_name.strip()
        if file_name:
            if file_name.count(":") > 1:
                # contains hg repo info
                sframe.location = basename(file_name.split(":")[-2])
            else:
                sframe.location = file_name
        elif lib_name:
            sframe.location = lib_name.strip()
        if line_no:
            sframe.offset = line_no.strip()
        elif offset:
            sframe.offset = offset.strip()
        return sframe


class GdbStackFrame(StackFrame):
    _re_gdb = re_compile(r"^#(?P<num>\d+)\s+(?P<off>0x[0-9a-f]+\sin\s)*(?P<line>.+)")

    @classmethod
    def from_line(cls, input_line: str) -> GdbStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            GdbStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        if "#" not in input_line:
            return None
        match = cls._re_gdb.match(input_line)
        if match is None:
            return None
        input_line = match.group("line").strip()
        if not input_line:
            return None
        sframe = cls(stack_line=match.group("num"))
        # sframe.offset = m.group("off")  # ignore binary offset for now
        # find function/method name
        match = _RE_FUNC_NAME.match(input_line)
        if match is not None:
            sframe.function = match.group("func")
        # find file name and line number
        if ") at " in input_line:
            input_line = input_line.split(") at ")[-1]
            with suppress(ValueError):
                input_line, sframe.offset = input_line.split(":")
            sframe.location = basename(input_line).split()[0]
        return sframe


class RrStackFrame(StackFrame):
    _re_rr = re_compile(r"rr\((?P<loc>.+)\+(?P<off>0x[0-9a-f]+)\)\[0x[0-9a-f]+\]")

    @classmethod
    def from_line(cls, input_line: str) -> RrStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            RrStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        if "rr(" not in input_line:
            return None
        match = cls._re_rr.match(input_line)
        if match is None:
            return None
        return cls(location=match.group("loc"), offset=match.group("off"))


class RustStackFrame(StackFrame):
    _re_rust_frame = re_compile(r"^\s+(?P<num>\d+):\s+0x[0-9a-f]+\s+\-\s+(?P<line>.+)")

    @classmethod
    def from_line(cls, input_line: str) -> RustStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            RustStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        match = cls._re_rust_frame.match(input_line)
        if match is None:
            return None
        sframe = cls(stack_line=match.group("num"))
        sframe.function = match.group("line").strip().rsplit("::h", 1)[0]
        return sframe


class SanitizerStackFrame(StackFrame):
    _re_sanitizer = re_compile(
        r"^\s*#(?P<num>\d+)\s0x[0-9a-f]+(?P<in>\sin)?\s+(?P<line>.+)"
    )

    @classmethod
    def from_line(cls, input_line: str) -> SanitizerStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            SanitizerStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        if "#" not in input_line:
            return None
        match = cls._re_sanitizer.match(input_line)
        if match is None:
            return None
        sframe = cls(stack_line=match.group("num"))
        input_line = match.group("line")
        # check if line is symbolized
        if match.group("in"):
            # find function/method name
            match = _RE_FUNC_NAME.match(input_line)
            if match:
                sframe.function = match.group("func")
                # remove function name
                input_line = input_line.split(" ", 1)[-1]
        if input_line.startswith("("):
            input_line = input_line.strip("()")
        # find location (file name or module) and offset (line # or offset)
        offset = re_match(r"(.+?)(\:([0-9a-f]+)|\+(0x[0-9a-f]+)).*", input_line)
        if offset:
            sframe.location = basename(offset.group(1))
            sframe.offset = offset.group(3) or offset.group(4)
        else:
            sframe.location = input_line
        return sframe


class ThreadSanitizerStackFrame(StackFrame):
    _re_tsan = re_compile(
        r"^\s*#(?P<num>\d+)\s(?P<line>.+)\s\(((?P<mod>.+)\+)?(?P<off>0x[0-9a-f]+)\)"
    )

    @classmethod
    def from_line(cls, input_line: str) -> ThreadSanitizerStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            ThreadSanitizerStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        if "#" not in input_line:
            return None
        match = cls._re_tsan.match(input_line)
        if match is None:
            return None
        sframe = cls(stack_line=match.group("num"))
        input_line = match.group("line")
        location_raw = basename(input_line)
        # try to parse file name and line number
        if location_raw:
            location_parts = location_raw.split()[-1].split(":")
            if location_parts and location_parts[0] != "<null>":
                sframe.location = location_parts.pop(0)
                if location_parts and location_parts[0] != "<null>":
                    sframe.offset = location_parts.pop(0)
        # use module name if file name cannot be found
        if not sframe.location:
            sframe.location = match.group("mod")
        # use module offset if line number cannot be found
        if not sframe.offset:
            sframe.offset = match.group("off")
        match = _RE_FUNC_NAME.match(input_line)
        if match is not None:
            function = match.group("func")
            if function and function != "<null>":
                sframe.function = function
        return sframe


class ValgrindStackFrame(StackFrame):
    _re_valgrind = re_compile(
        r"^==\d+==\s+(at|by)\s+0x[0-9A-F]+\:\s+(?P<func>.+?)\s+\((?P<line>.+)\)"
    )

    @classmethod
    def from_line(cls, input_line: str) -> ValgrindStackFrame | None:
        """Parse stack frame details.

        Args:
            input_line: A single line of text.

        Returns:
            ValgrindStackFrame
        """
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        if "== " not in input_line:
            return None
        match = cls._re_valgrind.match(input_line)
        if match is None:
            return None
        input_line = match.group("line")
        if input_line is None:  # pragma: no cover
            # this should not happen
            LOG.warning("failure in ValgrindStackFrame.from_line()")
            return None
        sframe = cls(function=match.group("func"))
        try:
            location, sframe.offset = input_line.split(":")
            sframe.location = location.strip()
        except ValueError:
            # trim anything from the beginning we might have missed
            location = input_line.rsplit("(")[-1]
            if location.startswith("in "):
                location = input_line[3:]
            sframe.location = basename(location)
        if not sframe.location:
            return None
        return sframe


class Stack:
    __slots__ = ("_height_limit", "_major", "_major_depth", "_minor", "frames")

    def __init__(
        self,
        frames: list[StackFrame],
        height_limit: int = 0,
        major_depth: int = MAJOR_DEPTH,
    ) -> None:
        assert height_limit >= 0
        assert major_depth >= 0
        self.frames = frames
        # use 0 for no limit
        self._height_limit = height_limit
        # use 0 for no limit for no limit
        self._major_depth = major_depth
        self._major: str | None = None
        self._minor: str | None = None

    def __str__(self) -> str:
        return "\n".join(str(frame) for frame in self.frames)

    def _calculate_hash(self, major: bool = False) -> str | None:
        """Calculate hash value from frames.

        Args:
            major: Perform major has calculation.

        Returns:
            Hash string.
        """
        if not self.frames or (major and self._major_depth < 1):
            return None

        if self._height_limit == 0:
            offset = 0
        else:
            offset = max(len(self.frames) - self._height_limit, 0)

        bucket_hash = sha1()
        major_depth = 0
        for frame in self.frames[offset:]:
            # only track depth when needed
            # and don't count ignored frames towards major hash depth
            if (major and self._major_depth > 0) and (
                not frame.function
                or not any(frame.function.startswith(x) for x in IGNORED_FRAMES)
            ):
                major_depth += 1
                if major_depth > self._major_depth:
                    break

            if frame.location is not None:
                bucket_hash.update(frame.location.encode(errors="replace"))
            if frame.function is not None:
                bucket_hash.update(frame.function.encode(errors="replace"))
            if major_depth > 1:
                # only add the offset from the top frame when calculating
                # the major hash and skip the rest
                continue
            if frame.offset is not None:
                bucket_hash.update(frame.offset.encode(errors="replace"))
        return bucket_hash.hexdigest()

    @classmethod
    def from_text(cls, input_text: str, major_depth: int = MAJOR_DEPTH) -> Stack:
        """Parse a stack trace from text. This is intended to parse the output
        from a single result. Some debuggers such as ASan and TSan can include
        multiple stacks per result.

        Args:
            input_text: Data to parse.
            major_depth: Number of frames to use to calculate the major hash. Use 0
                for no limit.

        Returns:
            Stack
        """

        frames: list[StackFrame] = []
        parser_class: type[StackFrame] | None = None
        for line in input_text.split("\n"):
            line = line.rstrip()
            if not line:
                # skip empty lines
                continue
            frame: StackFrame | None = None
            try:
                # only use a single StackFrame type
                if parser_class is None:
                    # order most to least common
                    for frame_parser in (
                        SanitizerStackFrame,
                        MinidumpStackFrame,
                        ThreadSanitizerStackFrame,
                        ValgrindStackFrame,
                        RrStackFrame,
                        RustStackFrame,
                        GdbStackFrame,
                    ):
                        frame = frame_parser.from_line(line)
                        if frame is not None:
                            LOG.debug("frame parser: %s", frame_parser.__name__)
                            parser_class = frame_parser
                            break
                else:
                    frame = parser_class.from_line(line)
            except Exception:  # pragma: no cover
                LOG.error("Error calling from_line() with: %r", line)
                raise
            if frame is None:
                continue

            if frame.stack_line is not None and frames:
                num = int(frame.stack_line)
                # check for new stack
                if num == 0:
                    # select stack to use
                    if parser_class in (SanitizerStackFrame, ThreadSanitizerStackFrame):
                        break
                    frames.clear()
                # check for out of order or missing frames
                elif frames[-1].stack_line and num - 1 != int(frames[-1].stack_line):
                    LOG.debug("scrambled logs?")
                    break
            frames.append(frame)

        return cls(frames, major_depth=major_depth)

    @property
    def height_limit(self) -> int:
        """Height limit used to calculate hash. The stack height is calculated from
        the entry point.

        Args:
            None

        Returns:
            Height limit.
        """
        return self._height_limit

    @height_limit.setter
    def height_limit(self, value: int) -> None:
        assert isinstance(value, int)
        assert value >= 0
        self._height_limit = value
        # force recalculation of hashes
        self._major = None
        self._minor = None

    @property
    def major(self) -> str | None:
        if self._major is None:
            self._major = self._calculate_hash(major=True)
        return self._major

    @property
    def minor(self) -> str | None:
        if self._minor is None:
            self._minor = self._calculate_hash()
        return self._minor


if __name__ == "__main__":
    from argparse import ArgumentParser, Namespace
    from pathlib import Path

    def main(args: Namespace) -> None:
        # set output verbosity
        if args.debug:
            basicConfig(format="[%(levelname).1s] %(message)s", level=DEBUG)
        else:
            basicConfig(format="%(message)s", level=INFO)

        with args.input.open("rb") as in_fp:
            stack = Stack.from_text(in_fp.read().decode(errors="replace"))
        for frame in stack.frames:
            LOG.info(frame)
        LOG.info("Minor: %s", stack.minor)
        LOG.info("Major: %s", stack.major)
        LOG.info("Frames: %d", len(stack.frames))

    parser = ArgumentParser()
    parser.add_argument("input", type=Path, help="File to scan for stack trace")
    parser.add_argument("-d", "--debug", action="store_true", help="Output debug info")
    main(parser.parse_args())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server
"""
from __future__ import annotations

from logging import getLogger
from pathlib import Path
from socket import SO_REUSEADDR, SOL_SOCKET, gethostname, socket
from ssl import PROTOCOL_TLS_SERVER, SSLContext, SSLSocket
from time import perf_counter, sleep
from typing import TYPE_CHECKING, Callable, cast

from .connection_manager import ConnectionManager
from .job import Job, Served

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Iterable, Mapping

    from .certificate_bundle import CertificateBundle
    from .server_map import ServerMap

__all__ = (
    "BLOCKED_PORTS",
    "Sapphire",
    "create_listening_socket",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


# collection of ports to avoid
# see: searchfox.org/mozilla-central/source/netwerk/base/nsIOService.cpp
# include ports above 1024
BLOCKED_PORTS = frozenset(
    (
        1719,
        1720,
        1723,
        2049,
        3659,
        4045,
        5060,
        5061,
        6000,
        6566,
        6665,
        6666,
        6667,
        6668,
        6669,
        6697,
        10080,
    )
)
LOG = getLogger(__name__)


def create_listening_socket(
    attempts: int = 10,
    port: int = 0,
    remote: bool = False,
) -> socket:
    """Create listening socket. Search for an open socket if needed and configure the
    socket. If the specified port is unavailable an OSError or PermissionError will be
    raised. If an available port cannot be found a RuntimeError will be raised.

    Args:
        attempts: Number of attempts to configure the socket.
        port: Port to listen on. Use 0 for system assigned port.
        remote: Accept all (non-local) incoming connections.

    Returns:
        A listening socket.
    """
    assert attempts > 0
    assert 0 <= port <= 65535

    if port in BLOCKED_PORTS or 0 < port <= 1024:
        raise ValueError("Cannot bind to blocked ports or ports <= 1024")

    for remaining in reversed(range(attempts)):
        sock = socket()
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        # attempt to bind/listen
        try:
            sock.bind(("0.0.0.0" if remote else "127.0.0.1", port))
            sock.listen(5)
            # put socket in non-blocking mode
            sock.settimeout(0)
        except (OSError, PermissionError) as exc:
            sock.close()
            if remaining > 0:
                LOG.debug("%s: %s", type(exc).__name__, exc)
                sleep(0.1)
                continue
            raise
        # avoid blocked ports
        if port == 0 and sock.getsockname()[1] in BLOCKED_PORTS:
            LOG.debug("bound to blocked port, retrying...")
            sock.close()
            continue
        # success
        break
    else:
        raise RuntimeError("Could not find available port")
    return sock


class Sapphire:
    __slots__ = ("_auto_close", "_max_workers", "_socket", "scheme", "timeout")

    def __init__(
        self,
        allow_remote: bool = False,
        auto_close: int = -1,
        certs: CertificateBundle | None = None,
        max_workers: int = 10,
        port: int = 0,
        timeout: int = 60,
    ) -> None:
        assert timeout >= 0
        self._auto_close = auto_close  # call 'window.close()' on 4xx error pages
        self._max_workers = max_workers  # limit worker threads
        sock = create_listening_socket(port=port, remote=allow_remote)
        # enable https if certificates are provided
        if certs:
            context = SSLContext(PROTOCOL_TLS_SERVER)
            context.load_cert_chain(certs.host, certs.key)
            self._socket: socket | SSLSocket = context.wrap_socket(
                sock, server_side=True
            )
            self.scheme = "https"
        else:
            self._socket = sock
            self.scheme = "http"
        self.timeout = timeout

    def __enter__(self) -> Sapphire:
        return self

    def __exit__(self, *exc: object) -> None:
        self.close()

    def clear_backlog(self, timeout: float = 10) -> bool:
        """Remove all pending connections from backlog. This should only be
        called when there isn't anything actively trying to connect.

        Args:
            timeout: Maximum number of seconds to run.

        Returns:
            True if all connections are cleared from the backlog otherwise False.
        """
        # this assumes the socket is in non-blocking mode
        assert not self._socket.getblocking()
        LOG.debug("clearing socket backlog")
        deadline = perf_counter() + timeout
        while True:
            try:
                self._socket.accept()[0].close()
            except BlockingIOError:
                # no remaining pending connections
                break
            except OSError as exc:
                LOG.debug("Error closing socket: %r", exc)
            else:
                LOG.debug("pending socket closed")
            # if this fires something is likely actively trying to connect
            if deadline <= perf_counter():
                return False
            # avoid hogging the cpu
            sleep(0.1)
        return True

    def close(self) -> None:
        """Close listening server socket.

        Args:
            None

        Returns:
            None
        """
        self._socket.close()

    @property
    def port(self) -> int:
        """Port number of listening socket.

        Args:
            None

        Returns:
            Listening port number.
        """
        return cast(int, self._socket.getsockname()[1])

    def serve_path(
        self,
        path: Path,
        continue_cb: Callable[[], bool] | None = None,
        forever: bool = False,
        required_files: Iterable[str] | None = None,
        server_map: ServerMap | None = None,
    ) -> tuple[Served, Mapping[str, Path]]:
        """Serve files in path.

        The status codes include:
            - Served.ALL: All required files were served
            - Served.NONE: No files were served
            - Served.REQUEST: Some files were requested

        Args:
            path: Directory to use as wwwroot.
            continue_cb: A callback that can be used to exit the serve loop.
                This must be a callable that returns a bool.
            forever: Continue to handle requests even after all files have been served.
                This is meant to be used with continue_cb.
            required_files: Files that need to be served in order to exit the
                serve loop.
            server_map: Map of server includes, dynamic requests and redirects.

        Returns:
            Status code and files served.
        """
        assert isinstance(path, Path)
        assert self.timeout >= 0
        LOG.debug("serving '%s' (forever=%r, timeout=%d)", path, forever, self.timeout)
        job = Job(
            path,
            auto_close=self._auto_close,
            forever=forever,
            required_files=required_files,
            server_map=server_map,
        )
        with ConnectionManager(job, self._socket, limit=self._max_workers) as mgr:
            timed_out = not mgr.serve(self.timeout, continue_cb=continue_cb)
        LOG.debug("%s, timed out: %r", job.status, timed_out)
        return (Served.TIMEOUT if timed_out else job.status, job.served)

    @classmethod
    def main(cls, args: Namespace) -> None:
        try:
            with cls(
                allow_remote=args.remote, port=args.port, timeout=args.timeout
            ) as serv:
                LOG.info(
                    "Serving '%s' @ http://%s:%d/",
                    args.path.resolve(),
                    gethostname() if args.remote else "127.0.0.1",
                    serv.port,
                )
                serv.serve_path(args.path, forever=True)
        except KeyboardInterrupt:
            LOG.warning("Ctrl+C detected. Shutting down...")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from enum import IntEnum, unique
from json import dumps, loads
from logging import getLogger
from os import getenv
from pathlib import Path
from shutil import copyfile, move, rmtree
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, cast, final
from zipfile import ZIP_DEFLATED, ZipFile

from Collector.Collector import Collector
from psutil import disk_usage

from .report import Report
from .utils import grz_tmp

if TYPE_CHECKING:
    from .storage import TestCase

__all__ = ("FilesystemReporter", "FuzzManagerReporter", "Quality")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


@unique
class Quality(IntEnum):
    """testcase quality values"""

    # final reduced testcase
    REDUCED = 0
    # original used for reduction (a reduced version exists)
    ORIGINAL = 1
    # used to manually mark a report as unreducible
    IGNORED = 3
    # the testcase is currently being reduced
    REDUCING = 4
    # haven't attempted reduction yet (1st attempt, generic reducer)
    UNREDUCED = 5
    # platform specific reduction requested (2nd attempt)
    REQUEST_SPECIFIC = 6
    # testcase not detected ("testcase" not a testcase?)
    NO_TESTCASE = 7
    # the testcase was reproducible, but broke during reduction
    REDUCER_BROKE = 8
    # reducer error
    REDUCER_ERROR = 9
    # could not reproduce the testcase
    NOT_REPRODUCIBLE = 10


class Reporter(metaclass=ABCMeta):
    __slots__ = ("display_logs",)

    def __init__(self) -> None:
        self.display_logs = getenv("GRZ_DISPLAY_REPORT") == "1"

    @abstractmethod
    def _post_submit(self) -> None:
        """Called after _submit_report() is called.

        Args:
            None

        Returns:
            None
        """

    @abstractmethod
    def _pre_submit(self, report: Report) -> None:
        """Called before _submit_report() is called.

        Args:
            report: Report to submit.

        Returns:
            None
        """

    @abstractmethod
    def _submit_report(
        self, report: Report, test_cases: list[TestCase], force: bool
    ) -> Any:
        """Implementation specific report submission code.

        Args:
            report: Report to submit.
            test_cases: Testcases to be submitted.
            force: Ignore any limits and always submit report.

        Returns:
            Implementation specific result indicating where the report was created.
        """

    @final
    def submit(
        self, test_cases: list[TestCase], report: Report, force: bool = False
    ) -> Any:
        """Submit report containing results.

        Args:
            test_cases: A collection of testcases, ordered oldest to newest,
                the newest being the mostly likely to trigger
                the result (crash, assert... etc).
            report: Report to submit.
            force: Ignore any limits and always submit report.

        Returns:
            Implementation specific result indicating where the report was created.
        """
        assert (
            not test_cases or test_cases[0].timestamp <= test_cases[-1].timestamp
        ), "tests must be ordered oldest to newest"
        self._pre_submit(report)
        # output report contents to console
        if self.display_logs:
            if not report.is_hang:
                assert report.preferred
                LOG.info(
                    "=== BEGIN REPORT ===\n%s",
                    report.preferred.read_text("utf-8", errors="ignore"),
                )
            else:
                LOG.info("=== BEGIN REPORT ===\nBrowser hang detected")
            LOG.info("=== END REPORT ===")
        result = self._submit_report(report, test_cases, force)
        if report is not None:
            report.cleanup()
        self._post_submit()
        return result


class FilesystemReporter(Reporter):
    DISK_SPACE_ABORT = 512 * 1024 * 1024  # 512 MB

    __slots__ = ("major_bucket", "min_space", "report_path", "report_prefix")

    def __init__(self, report_path: Path, major_bucket: bool = True) -> None:
        super().__init__()
        self.major_bucket = major_bucket
        self.min_space = FilesystemReporter.DISK_SPACE_ABORT
        self.report_path = report_path
        self.report_prefix: str | None = None

    def _pre_submit(self, report: Report) -> None:
        self.report_prefix = report.prefix

    def _post_submit(self) -> None:
        pass

    def _submit_report(
        self, report: Report, test_cases: list[TestCase], force: bool
    ) -> Path:
        # create major bucket directory in working directory if needed
        if self.major_bucket:
            dest = self.report_path / report.major[:16]
        else:
            dest = self.report_path
        dest.mkdir(parents=True, exist_ok=True)
        # dump test cases and the contained files to working directory
        for test_number, test_case in enumerate(reversed(test_cases)):
            dump_path = dest / f"{self.report_prefix}-{test_number}"
            dump_path.mkdir(exist_ok=True)
            test_case.dump(dump_path, include_details=True)
        # move logs into bucket directory
        log_path = dest / f"{self.report_prefix}_logs"
        if log_path.is_dir():
            LOG.warning("Report log path exists '%s'", log_path)
        move(report.path, log_path)
        # avoid filling the disk
        free_space = disk_usage(str(log_path.resolve())).free
        if free_space < self.min_space:
            raise RuntimeError(f"Low disk space: {free_space / 1_048_576:0.1f}MBs")
        return dest


class FailedLaunchReporter(FilesystemReporter):
    """Save launch failure reports to disk."""

    def __init__(self, display: bool = False) -> None:
        super().__init__(grz_tmp("launch_failures"), major_bucket=False)
        self.display_logs = display
        if not display:
            LOG.info("Display logs with --display-launch-failures")

    def _post_submit(self) -> None:
        super()._post_submit()
        LOG.info(
            "Logs for %r can be found here '%s'", self.report_prefix, self.report_path
        )


class FuzzManagerReporter(Reporter):
    __slots__ = ("_extra_metadata", "tool")

    def __init__(self, tool: str) -> None:
        super().__init__()
        self._extra_metadata: dict[str, Any] = {}
        # remove whitespace and use only lowercase
        self.tool = "-".join(tool.lower().split())
        assert self.tool, "tool value cannot be empty"

    def _post_submit(self) -> None:
        self._extra_metadata.clear()

    def add_extra_metadata(self, key: str, value: Any) -> None:
        """Add extra metadata to be reported with any CrashEntry objects reported.

        Arguments:
            key: key for this data in the metadata dict.
            value: JSON serializable object to be included in the FM crash
                   metadata. The object will be deep-copied.

        Returns:
            None
        """
        assert isinstance(key, str)
        assert key not in self._extra_metadata
        # deep copy and ensure that value is JSON serializable
        self._extra_metadata[key] = loads(dumps(value))

    def _pre_submit(self, report: Report) -> None:
        self._process_rr_trace(report)

    def _process_rr_trace(self, report: Report) -> None:
        # don't report large files to FuzzManager
        trace_path = report.path / "rr-traces"
        if trace_path.is_dir():
            LOG.info("Ignored rr trace")
            self.add_extra_metadata("rr-trace", "ignored")
            # remove traces so they are not uploaded to FM (because they are huge)
            rmtree(trace_path)

    def _submit_report(
        self, report: Report, test_cases: list[TestCase], force: bool
    ) -> int | None:
        collector = Collector(tool=self.tool)

        if not force:
            if collector.sigCacheDir and Path(collector.sigCacheDir).is_dir():
                # search for a cached signature match
                _, cache_metadata = collector.search(report.crash_info)
                # check if signature has been marked as frequent in FM
                if cache_metadata and cache_metadata["frequent"]:
                    LOG.info(
                        "Frequent crash matched existing signature: %s",
                        cache_metadata["shortDescription"],
                    )
                    return None
            else:
                LOG.debug("sigCacheDir does not exist (%r)", collector.sigCacheDir)

        if report.is_hang:
            self.add_extra_metadata("is_hang", True)

        if report.unstable:
            self.add_extra_metadata("unstable_build", True)

        if test_cases:
            self.add_extra_metadata("testcase_count", len(test_cases))
            # dump test cases and the contained files to working directory
            for test_number, test_case in enumerate(reversed(test_cases)):
                dump_path = report.path / f"{report.prefix}-{test_number}"
                dump_path.mkdir(exist_ok=True)
                test_case.dump(dump_path, include_details=True)
            quality = Quality.UNREDUCED
            if test_cases[0].env_vars:
                # add environment variables to metadata
                report.crash_info.configuration.addEnvironmentVariables(
                    test_cases[0].env_vars
                )
        else:
            quality = Quality.NO_TESTCASE
        report.crash_info.configuration.addMetadata(self._extra_metadata)

        # grab screen log (used in automation)
        if getenv("WINDOW") is not None:
            screen_log = Path.cwd() / f"screenlog.{getenv('WINDOW')}"
            if screen_log.is_file():
                target_log = report.path / "screenlog.txt"
                copyfile(screen_log, target_log)
                Report.tail(target_log, 10240)  # limit to last 10K

        with TemporaryDirectory(prefix="fm-zip", dir=grz_tmp()) as tmp_dir:
            # add results to a zip file
            zip_name = Path(tmp_dir) / f"{report.prefix}.zip"
            with ZipFile(zip_name, mode="w", compression=ZIP_DEFLATED) as zip_fp:
                # add test files
                for entry in report.path.rglob("*"):
                    if entry.is_file():
                        zip_fp.write(entry, arcname=entry.relative_to(report.path))

            # submit results to the FuzzManager server
            new_entry = collector.submit(
                report.crash_info, testCase=zip_name, testCaseQuality=quality.value
            )
        LOG.info("Reported: %d, %s, %s", new_entry["id"], quality.name, collector.tool)
        return cast(int, new_entry["id"])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

from json import dumps, loads

from pytest import mark, raises

from ..target import AssetManager
from .storage import TEST_INFO, TestCase, TestCaseLoadFailure, TestFileExists


def test_testcase_01(tmp_path):
    """test empty TestCase"""
    entry_point = "test.html"
    adpt_name = "test-adapter"
    with TestCase(entry_point, adpt_name) as tcase:
        assert tcase.entry_point == entry_point
        assert entry_point not in tcase
        assert not tcase.assets
        assert not tcase.assets_path
        assert tcase.adapter_name == adpt_name
        assert tcase.duration is None
        assert tcase.data_size == 0
        assert not tcase.https
        assert tcase.input_fname is None
        assert tcase.timestamp > 0
        assert not tcase.env_vars
        assert tcase.root
        assert not tcase._files.optional
        assert not tcase._files.required
        assert not tcase._in_place
        assert len(tcase) == 0
        assert not any(tcase)
        assert not any(tcase.optional)
        assert not any(tcase.required)
        tcase.dump(tmp_path)
        assert not any(tmp_path.iterdir())
        tcase.dump(tmp_path, include_details=True)
        assert (tmp_path / TEST_INFO).is_file()
        tcase.cleanup()
        assert not tcase.root.is_dir()


def test_testcase_02():
    """test TestCase.add_from_file() - add files with existing, in place data"""
    with TestCase("test.html", "adpt") as tcase:
        (tcase.root / "test.html").touch()
        (tcase.root / "opt.html").touch()
        tcase.add_from_file(tcase.root / "test.html", required=True)
        assert len(tcase) == 1
        assert "test.html" in tcase
        assert len(tuple(tcase.required)) == 1
        assert not any(tcase.optional)
        tcase.add_from_file(tcase.root / "opt.html", required=False)
        assert len(tcase) == 2
        assert "opt.html" in tcase
        assert len(tuple(tcase.required)) == 1
        assert len(tuple(tcase.optional)) == 1
        # add previously added file
        with raises(TestFileExists, match="'opt.html' exists in test"):
            tcase.add_from_file(tcase.root / "opt.html", required=False)


@mark.parametrize(
    "copy, required",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_testcase_03(tmp_path, copy, required):
    """test TestCase.add_from_file()"""
    in_file = tmp_path / "file.bin"
    in_file.write_text("data")
    with TestCase("test.html", "adpt", input_fname="in.bin") as tcase:
        tcase.add_from_file(in_file, copy=copy, required=required)
        assert tcase.data_size == 4
        assert "file.bin" in tcase
        assert tcase["file.bin"].is_file()
        if required:
            assert in_file.name not in tcase.optional
        else:
            assert in_file.name in tcase.optional
        assert in_file.exists() == copy
        # try overwriting existing file
        if copy:
            with raises(TestFileExists, match="'file.bin' exists in test"):
                tcase.add_from_file(in_file, copy=True)
            assert in_file.exists()
        else:
            assert not in_file.exists()


@mark.parametrize(
    "file_paths",
    [
        ("a.bin",),
        ("a/a.bin",),
        ("a.bin", "b.bin"),
        ("a.bin", "b/c.bin", "b/d.bin"),
    ],
)
def test_testcase_04(tmp_path, file_paths):
    """test TestCase.add_from_file()"""
    with TestCase("test.html", "adpt") as tcase:
        for file_path in file_paths:
            src_file = tmp_path / file_path
            src_file.parent.mkdir(exist_ok=True, parents=True)
            src_file.write_text("data")
            tcase.add_from_file(src_file, file_name=file_path, required=True)
            assert file_path in tcase
            assert file_path in tcase.required
            assert file_path not in tcase.optional
            assert (tcase.root / file_path).is_file()


def test_testcase_05():
    """test TestCase.add_from_bytes()"""
    with TestCase("a.html", "adpt") as tcase:
        tcase.add_from_bytes(b"foo", "a.html", required=True)
        tcase.add_from_bytes(b"foo", "b.html", required=False)
        assert "a.html" in tcase._files.required
        assert "b.html" in tcase._files.optional
        # add file with invalid file name
        with raises(ValueError, match="invalid path ''"):
            tcase.add_from_bytes(b"foo", "", required=False)


def test_testcase_06():
    """test TestCase.data_size"""
    with TestCase("test.html", "test-adapter") as tcase:
        assert tcase.data_size == 0
        tcase.add_from_bytes(b"1", "testfile1.bin", required=True)
        assert tcase.data_size == 1
        tcase.add_from_bytes(b"12", "testfile2.bin", required=False)
        assert tcase.data_size == 3


def test_testcase_07(tmp_path):
    """test TestCase.read_info()"""
    # missing test info file
    assert not TestCase.read_info(tmp_path)
    # invalid test info file
    (tmp_path / TEST_INFO).write_text("X")
    with raises(TestCaseLoadFailure, match=f"Invalid '{TEST_INFO}'"):
        TestCase.read_info(tmp_path)
    # test info file missing 'target' entry
    (tmp_path / TEST_INFO).write_text("{}")
    with raises(TestCaseLoadFailure, match=f"Invalid 'target' entry in '{TEST_INFO}'"):
        TestCase.read_info(tmp_path)
    # success
    (tmp_path / TEST_INFO).write_text('{"target": "foo"}')
    assert TestCase.read_info(tmp_path) == {"target": "foo"}


def test_testcase_08(tmp_path):
    """test TestCase._find_entry_point()"""
    # empty directory
    with raises(TestCaseLoadFailure, match="Could not determine entry point"):
        TestCase._find_entry_point(tmp_path)
    # missing potential entry point
    (tmp_path / TEST_INFO).touch()
    with raises(TestCaseLoadFailure, match="Could not determine entry point"):
        TestCase._find_entry_point(tmp_path)
    # success
    (tmp_path / "poc.html").touch()
    assert TestCase._find_entry_point(tmp_path) == (tmp_path / "poc.html")
    # Ambiguous entry point
    (tmp_path / "other.html").touch()
    with raises(TestCaseLoadFailure, match="Ambiguous entry point"):
        TestCase._find_entry_point(tmp_path)


def test_testcase_09(tmp_path):
    """test TestCase.load_meta()"""
    # empty directory
    with raises(TestCaseLoadFailure, match="Could not determine entry point"):
        TestCase.load_meta(tmp_path)
    # missing directory
    with raises(TestCaseLoadFailure, match="Missing or invalid TestCase"):
        TestCase.load_meta(tmp_path / "missing")
    # success (directory)
    (tmp_path / "test_01.html").touch()
    entry_point, info = TestCase.load_meta(tmp_path)
    assert entry_point == tmp_path / "test_01.html"
    assert not info
    # success (file)
    entry_point, info = TestCase.load_meta(tmp_path / "test_01.html")
    assert entry_point == tmp_path / "test_01.html"
    assert not info
    # success (with test info file)
    (tmp_path / TEST_INFO).write_text('{"target": "test_01.html"}')
    (tmp_path / "other.html").touch()
    entry_point, info = TestCase.load_meta(tmp_path)
    assert entry_point == (tmp_path / "test_01.html")
    assert info.get("target") == "test_01.html"
    # success (with test info file) override entry point
    entry_point, info = TestCase.load_meta(
        tmp_path, entry_point=(tmp_path / "other.html")
    )
    assert entry_point == tmp_path / "other.html"
    assert info.get("target") == "other.html"
    # invalid test info file (will fallback to searching for test)
    (tmp_path / "other.html").unlink()
    (tmp_path / TEST_INFO).write_text("{}")
    entry_point, info = TestCase.load_meta(tmp_path)
    assert entry_point == (tmp_path / "test_01.html")
    assert not info


def test_testcase_10(tmp_path):
    """test TestCase.load()"""
    data = tmp_path / "test-data"
    data.mkdir()
    # empty directory
    with raises(TestCaseLoadFailure, match="Could not determine entry point"):
        TestCase.load(tmp_path)
    # missing directory
    with raises(TestCaseLoadFailure, match="Missing or invalid TestCase"):
        TestCase.load(tmp_path / "missing")
    # directory with test case
    (data / "poc.html").touch()
    loaded = TestCase.load(data)
    assert loaded._in_place
    assert loaded.entry_point == "poc.html"
    assert loaded.root == data
    # directory with test case invalid entry point
    (tmp_path / "external.html").touch()
    with raises(TestCaseLoadFailure, match="Entry point must be in root of given path"):
        TestCase.load(data, entry_point=tmp_path / "external.html")
    # single file directory
    loaded = TestCase.load(data / "poc.html")
    assert loaded._in_place
    assert loaded.entry_point == "poc.html"
    assert loaded.root == data


def test_testcase_11(tmp_path):
    """test TestCase.load() existing test case with simple test info file"""
    # build a test case
    src = tmp_path / "src"
    with TestCase("test.html", "test-adapter") as test:
        test.add_from_bytes(b"", test.entry_point, required=True)
        test.dump(src, include_details=True)
    # successful load
    loaded = TestCase.load(src)
    assert loaded._in_place
    assert loaded.entry_point == "test.html"
    assert loaded.root.samefile(src)
    assert len(tuple(loaded.required)) == 1
    assert not any(loaded.optional)


@mark.parametrize("catalog", [False, True])
def test_testcase_12(tmp_path, catalog):
    """test TestCase.load() existing test case with test info file"""
    # build a test case
    asset_file = tmp_path / "asset.txt"
    asset_file.touch()
    src = tmp_path / "src"
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("example", asset_file)
        with TestCase("test.html", "test-adapter") as test:
            test.add_from_bytes(b"", test.entry_point, required=True)
            test.add_from_bytes(b"", "optional.bin", required=False)
            test.add_from_bytes(b"", "nested/a.html", required=False)
            test.assets = dict(asset_mgr.assets)
            test.assets_path = asset_mgr.path
            test.env_vars["TEST_ENV_VAR"] = "100"
            test.dump(src, include_details=True)
    # successful load
    loaded = TestCase.load(src, catalog=catalog)
    assert loaded.entry_point == "test.html"
    assert loaded.root.samefile(src)
    assert (loaded.root / "optional.bin").is_file()
    assert (loaded.root / "nested" / "a.html").is_file()
    if catalog:
        assert "optional.bin" in loaded.optional
        assert "nested/a.html" in loaded.optional
        assert "_assets_/asset.txt" not in loaded.optional
        assert TEST_INFO not in loaded.optional
    else:
        assert not any(loaded.optional)
    assert loaded.assets == {"example": "asset.txt"}
    assert loaded.assets_path is not None
    assert (loaded.root / loaded.assets_path / "asset.txt").is_file()
    assert loaded.env_vars.get("TEST_ENV_VAR") == "100"
    assert len(tuple(loaded.required)) == 1
    # this should do nothing
    loaded.cleanup()
    assert loaded.root.is_dir()


def test_testcase_13(tmp_path):
    """test TestCase.load() test info file error cases"""
    # bad 'assets' entry in test info file
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    entry_point = src_dir / "target.html"
    entry_point.touch()
    with TestCase("target.html", "test-adapter") as src:
        src.dump(src_dir, include_details=True)
    test_info = loads((src_dir / TEST_INFO).read_text())
    test_info["assets"] = {"bad": 1}
    (src_dir / TEST_INFO).write_text(dumps(test_info))
    with raises(TestCaseLoadFailure, match="'assets' contains invalid entry"):
        TestCase.load(src_dir)
    # bad 'env' entry in test info file
    with TestCase("target.html", "test-adapter") as src:
        src.dump(src_dir, include_details=True)
    test_info = loads((src_dir / TEST_INFO).read_text())
    test_info["env"] = {"bad": 1}
    (src_dir / TEST_INFO).write_text(dumps(test_info))
    with raises(TestCaseLoadFailure, match="'env' contains invalid entry"):
        TestCase.load(src_dir)
    # missing asset data
    test_info["env"].clear()
    test_info["assets"] = {"a": "a"}
    test_info["assets_path"] = "missing"
    (src_dir / TEST_INFO).write_text(dumps(test_info))
    with TestCase.load(src_dir) as loaded:
        assert not loaded.assets
        assert loaded.assets_path is None


def test_testcase_14(tmp_path):
    """test TestCase.load() and TestCase.dump() with assets"""
    # build a test case
    asset_file = tmp_path / "asset.txt"
    asset_file.touch()
    src = tmp_path / "src"
    with AssetManager() as asset_mgr:
        asset_mgr.add("example", asset_file)
        with TestCase("test.html", "test-adapter") as test:
            test.add_from_bytes(b"", test.entry_point, required=True)
            test.assets = dict(asset_mgr.assets)
            test.assets_path = asset_mgr.path
            test.dump(src, include_details=True)
    # load
    loaded = TestCase.load(src)
    assert loaded.assets
    assert loaded.assets_path
    with AssetManager.load(loaded.assets, loaded.assets_path) as asset_mgr:
        # dump the AssetManager's assets (simulate how this works in replay)
        loaded.assets = dict(asset_mgr.assets)
        loaded.assets_path = asset_mgr.path
        # dump loaded test case
        dst = tmp_path / "dst"
        loaded.dump(dst, include_details=True)
    assert (dst / "test.html").is_file()
    assert (dst / "_assets_" / "asset.txt").is_file()


@mark.parametrize(
    "duration, hang, https, input_fname",
    [
        (1.2, True, True, "foo/test.name"),
        (0, False, False, None),
    ],
)
def test_testcase_15(tmp_path, duration, hang, https, input_fname):
    """test TestCase - dump, load and compare"""
    asset_path = tmp_path / "assets"
    asset_path.mkdir()
    asset = asset_path / "asset_file.txt"
    asset.touch()
    working = tmp_path / "working"
    with TestCase("a.html", "adpt") as org:
        # set non default values
        org.duration = duration
        org.env_vars = {"en1": "1", "en2": "2"}
        org.https = https
        org.hang = hang
        org.input_fname = input_fname
        org.add_from_bytes(b"a", org.entry_point)
        org.assets = {"sample": asset.name}
        org.assets_path = asset_path
        org.version = "1.2.3"
        org.dump(working, include_details=True)
        assert (working / "_assets_" / asset.name).is_file()
        with TestCase.load(working) as loaded:
            for prop in TestCase.__slots__:
                if not prop.startswith("_") and prop != "assets_path":
                    assert getattr(loaded, prop) == getattr(org, prop)
            assert not set(org) ^ set(loaded)
            assert "sample" in loaded.assets
            assert loaded.assets["sample"] == asset.name


def test_testcase_16():
    """test TestCase.__getitem__()"""
    with TestCase("test.htm", "test-adapter") as src:
        src.add_from_bytes(b"test", src.entry_point)
        with raises(KeyError, match="missing"):
            src["missing"]  # pylint: disable=pointless-statement
        assert src["test.htm"]


@mark.parametrize(
    "remote_assets",
    [
        # No assets
        None,
        # Remote assets
        True,
        # Local assets (assets only exist in test case)
        False,
    ],
)
def test_testcase_17(tmp_path, remote_assets):
    """test TestCase.clone()"""
    with TestCase("test.htm", "adpt", input_fname="fn") as src:
        if remote_assets:
            src.assets = {"foo": "asset.file"}
            src.assets_path = tmp_path
            (tmp_path / "asset.file").touch()
        elif remote_assets is not None:
            src.assets = {"foo": "asset.file"}
            src.assets_path = src.root / "_assets_"
            src.assets_path.mkdir()
            (src.assets_path / "asset.file").touch()
        src.duration = 1.2
        src.hang = True
        src.https = True
        src.add_from_bytes(b"123", "test.htm", required=True)
        src.add_from_bytes(b"456", "opt.htm", required=False)
        src.env_vars["foo"] = "bar"
        with src.clone() as dst:
            for prop in TestCase.__slots__:
                if prop.startswith("_"):
                    continue
                if remote_assets is False and prop == "assets_path":
                    assert src.assets_path != dst.assets_path
                    continue
                assert getattr(src, prop) == getattr(dst, prop)
            assert src.data_size == dst.data_size
            for file, data in (
                ("test.htm", b"123"),
                ("opt.htm", b"456"),
            ):
                assert src[file].read_bytes() == data
                assert dst[file].read_bytes() == data
                assert not dst[file].samefile(src[file])
            assert dst.env_vars == {"foo": "bar"}
            assert not set(src.optional) ^ set(dst.optional)
            assert not set(src.required) ^ set(dst.required)
            if remote_assets is not None:
                assert dst.assets_path
                assert (dst.assets_path / dst.assets["foo"]).is_file()


@mark.parametrize(
    "path",
    [
        # empty file name
        "",
        # path (root) with missing file name
        "/",
        # path (root) with missing file name
        "/.",
        # path with missing file name
        "path/",
        # outside of wwwroot
        "../test",
        # outside of wwwroot
        "a/../../b",
        # outside of wwwroot
        "..",
        # outside of wwwroot
        "C:\\test",
        # cwd
        ".",
    ],
)
def test_testcase_18(path):
    """test TestCase.sanitize_path() with invalid paths"""
    with raises(ValueError, match="invalid path"):
        TestCase.sanitize_path(path)


@mark.parametrize(
    "path, expected_result",
    [
        ("a", "a"),
        ("file.bin", "file.bin"),
        ("a/file.bin", "a/file.bin"),
        ("/file.bin", "file.bin"),
        ("./file.bin", "file.bin"),
        ("path\\file", "path/file"),
        ("\\\\a\\b\\file.bin", "a/b/file.bin"),
        (".file", ".file"),
        ("a/../file.bin", "file.bin"),
        ("./a/./b/../c", "a/c"),
    ],
)
def test_testcase_19(path, expected_result):
    """test TestCase.sanitize_path()"""
    assert TestCase.sanitize_path(path) == expected_result
